# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
#
# Source Attribution:
# The majority of this code is derived from the following sources:
# - TextAttack GitHub Repository: https://github.com/QData/TextAttack
# - Reference Paper: Morris, John X., et al. "TextAttack: A Framework for Adversarial Attacks, Data Augmentation, and Adversarial Training in NLP." arXiv preprint arXiv:2005.05909 (2020).


from abc import ABC, abstractmethod

import lru
import numpy as np
import torch

from textattack.goal_function_results.goal_function_result import (
    GoalFunctionResultStatus,
)
from textattack.shared.utils import ReprMixin
from textattack.goal_function_results import ClassificationGoalFunctionResult


class GoalFunction(ReprMixin, ABC):
    """Evaluates how well a perturbed attacked_text object is achieving a
    specified goal.

    Args:
        model_wrapper (:class:`~textattack.models.wrappers.ModelWrapper`):
            The victim model to attack.
        maximizable(:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether the goal function is maximizable, as opposed to a boolean result of success or failure.
        query_budget (:obj:`float`, `optional`, defaults to :obj:`float("in")`):
            The maximum number of model queries allowed.
        model_cache_size (:obj:`int`, `optional`, defaults to :obj:`2**20`):
            The maximum number of items to keep in the model results cache at once.
    """

    def __init__(
        self,
        model_wrapper,
        maximizable=False,
        use_cache=True,
        query_budget=float("inf"),
        model_batch_size=32,
        model_cache_size=2**20,
    ):
        self.model = model_wrapper
        self.maximizable = maximizable
        self.use_cache = use_cache
        self.query_budget = query_budget
        self.batch_size = model_batch_size
        self.ground_truth_output = -1
        if self.use_cache:
            self._call_model_cache = lru.LRU(model_cache_size)
        else:
            self._call_model_cache = None

    def clear_cache(self):
        if self.use_cache:
            self._call_model_cache.clear()

    def init_attack_example(self, attacked_text):
        """Called before attacking ``attacked_text`` to 'reset' the goal
        function and set properties for this example."""
        self.initial_attacked_text = attacked_text
        self.num_queries = 0
        result, _ = self.get_result(attacked_text, check_skip=True)
        self.ground_truth_output = result.output
        return result, _

    def get_output(self, attacked_text):
        """Returns output for display based on the result of calling the
        model."""
        return self._get_displayed_output(self._call_model([attacked_text])[0])

    def get_result(self, attacked_text, **kwargs):
        """A helper method that queries ``self.get_results`` with a single
        ``AttackedText`` object."""
        results, search_over = self.get_results([attacked_text], **kwargs)
        result = results[0] if len(results) else None
        return result, search_over

    def get_results(self, attacked_text_list, check_skip=False):
        """For each attacked_text object in attacked_text_list, returns a
        result consisting of whether or not the goal has been achieved, the
        output for display purposes, and a score.

        Additionally returns whether the search is over due to the query
        budget.
        """
        results = []
        if self.query_budget < float("inf"):
            queries_left = self.query_budget - self.num_queries
            attacked_text_list = attacked_text_list[:queries_left]
        self.num_queries += len(attacked_text_list)
        model_outputs = self._call_model(attacked_text_list)
        for attacked_text, raw_output in zip(attacked_text_list, model_outputs):
            displayed_output = self._get_displayed_output(raw_output)
            goal_status = self._get_goal_status(
                raw_output, attacked_text, check_skip=check_skip
            )
            goal_function_score = self._get_score(raw_output, attacked_text)
            results.append(
                self._goal_function_result_type()(
                    attacked_text,
                    raw_output,
                    displayed_output,
                    goal_status,
                    goal_function_score,
                    self.num_queries,
                    self.ground_truth_output,
                )
            )
        return results, self.num_queries == self.query_budget

    def _get_goal_status(self, model_output, attacked_text, check_skip=False):
        should_skip = check_skip and self._should_skip(model_output, attacked_text)
        if should_skip:
            return GoalFunctionResultStatus.SKIPPED
        if self.maximizable:
            return GoalFunctionResultStatus.MAXIMIZING
        if self._is_goal_complete(model_output, attacked_text):
            return GoalFunctionResultStatus.SUCCEEDED
        return GoalFunctionResultStatus.SEARCHING

    @abstractmethod
    def _is_goal_complete(self, model_output, attacked_text):
        raise NotImplementedError()

    def _should_skip(self, model_output, attacked_text):
        return self._is_goal_complete(model_output, attacked_text)

    @abstractmethod
    def _get_score(self, model_output, attacked_text):
        raise NotImplementedError()

    def _get_displayed_output(self, raw_output):
        return raw_output

    @abstractmethod
    def _goal_function_result_type(self):
        """Returns the class of this goal function's results."""
        raise NotImplementedError()

    @abstractmethod
    def _process_model_outputs(self, inputs, outputs):
        """Processes and validates a list of model outputs.

        This is a task-dependent operation. For example, classification
        outputs need to make sure they have a softmax applied.
        """
        raise NotImplementedError()

    def _call_model_uncached(self, attacked_text_list):
        """Queries model and returns outputs for a list of AttackedText
        objects."""
        if not len(attacked_text_list):
            return []

        inputs = [at.tokenizer_input for at in attacked_text_list]
        outputs = []
        i = 0
        while i < len(inputs):
            batch = inputs[i : i + self.batch_size]
            batch_preds = self.model(batch)

            # Some seq-to-seq models will return a single string as a prediction
            # for a single-string list. Wrap these in a list.
            if isinstance(batch_preds, str):
                batch_preds = [batch_preds]

            # Get PyTorch tensors off of other devices.
            if isinstance(batch_preds, torch.Tensor):
                batch_preds = batch_preds.cpu()

            if isinstance(batch_preds, list):
                outputs.extend(batch_preds)
            elif isinstance(batch_preds, np.ndarray):
                outputs.append(torch.tensor(batch_preds))
            else:
                outputs.append(batch_preds)
            i += self.batch_size

        if isinstance(outputs[0], torch.Tensor):
            outputs = torch.cat(outputs, dim=0)

        assert len(inputs) == len(
            outputs
        ), f"Got {len(outputs)} outputs for {len(inputs)} inputs"

        return self._process_model_outputs(attacked_text_list, outputs)

    def _call_model(self, attacked_text_list):
        """Gets predictions for a list of ``AttackedText`` objects.

        Gets prediction from cache if possible. If prediction is not in
        the cache, queries model and stores prediction in cache.
        """
        if not self.use_cache:
            return self._call_model_uncached(attacked_text_list)
        else:
            uncached_list = []
            for text in attacked_text_list:
                if text in self._call_model_cache:
                    # Re-write value in cache. This moves the key to the top of the
                    # LRU cache and prevents the unlikely event that the text
                    # is overwritten when we store the inputs from `uncached_list`.
                    self._call_model_cache[text] = self._call_model_cache[text]
                else:
                    uncached_list.append(text)
            uncached_list = [
                text
                for text in attacked_text_list
                if text not in self._call_model_cache
            ]
            outputs = self._call_model_uncached(uncached_list)
            for text, output in zip(uncached_list, outputs):
                self._call_model_cache[text] = output
            all_outputs = [self._call_model_cache[text] for text in attacked_text_list]
            return all_outputs

    def extra_repr_keys(self):
        attrs = []
        if self.query_budget < float("inf"):
            attrs.append("query_budget")
        if self.maximizable:
            attrs.append("maximizable")
        return attrs

    def __getstate__(self):
        state = self.__dict__.copy()
        if self.use_cache:
            state["_call_model_cache"] = self._call_model_cache.get_size()
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        if self.use_cache:
            self._call_model_cache = lru.LRU(state["_call_model_cache"])


class AdvPromptGoalFunction(GoalFunction):

    """A goal function defined on a model that outputs a probability for some
    number of classes."""

    def __init__(self, model, dataset, eval_func, query_budget, *args, target_max_acc=0, verbose=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        self.dataset = dataset
        self.eval_func = eval_func
        self.query_budget = query_budget
        self.target_max_acc = target_max_acc
        self.verbose = verbose

    def _call_model(self, attacked_text_list):
        acc_list = []
        for attacked_prompt in attacked_text_list:
            prompt = attacked_prompt.text
            
            acc_list.append(self.eval_func(prompt, self.dataset, self.model))
            if self.verbose:
                print("--------------------------------------------------")
                print("Current prompt is: ", prompt)
                print("Current accuracy is: ", acc_list[-1])
                print("--------------------------------------------------\n")

        return acc_list

    def _process_model_outputs(self, inputs, outputs):
        return outputs

    def _is_goal_complete(self, acc, _):
        return acc < self.target_max_acc
        
    def _get_score(self, model_output, _):
        score = self.ground_truth_output - model_output

        return score

    def _goal_function_result_type(self):
        """Returns the class of this goal function's results."""
        return ClassificationGoalFunctionResult

    def extra_repr_keys(self):
        return []

    def _get_displayed_output(self, raw_output):
        return raw_output

    def _get_goal_status(self, model_output, attacked_text, check_skip=False):
        return super()._get_goal_status(model_output, attacked_text, check_skip)


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import re
import os
from typing import Union, List


class Prompt:
    def __init__(self, prompt_content: Union[List[str], str, None] = None, dataset_name: str = None):
        
        if prompt_content:
            if isinstance(prompt_content, list):
                self.prompts = prompt_content
            else:
                self.prompts = [prompt_content]
        elif dataset_name:
            self.prompts = self._load_default_prompt(dataset_name)
            self.dataset_name = dataset_name
        else:
            raise ValueError("Either provide prompt_content or specify dataset_name, task_type, and shot_type.")
    
    def __len__(self):
        return len(self.prompts)

    def __getitem__(self, index):
        return self.prompts[index]
    
    def _load_default_prompt(self, dataset_name):
        from .task_oriented import TASK_ORIENTED_PROMPTS
        from .role_oriented import ROLE_ORIENTED_PROMPTS
        return TASK_ORIENTED_PROMPTS[dataset_name] + ROLE_ORIENTED_PROMPTS[dataset_name]

    def load_adv_prompt(self, model_name, dataset_name, attack_name, prompt_type):
        prompts_dict = retrieve(model_name, dataset_name, attack_name, prompt_type)
        self.prompts = [adv_prompt["attack prompt"] for adv_prompt in prompts_dict]
    
    def add_few_shot_examples(self, few_shot_examples=None):
        if few_shot_examples != None:
            few_shot_examples = self._load_default_few_shot_examples()
        
        self.prompts = [prompt + "\n" + few_shot_examples for prompt in self.prompts]

    def _load_default_few_shot_examples(self):
        cur_dir = os.path.dirname(os.path.abspath(__file__))
        few_shot_path = os.path.join(cur_dir, "few_shot_examples.yaml")
        import yaml
        few_shot_examples = yaml.load(open(few_shot_path, "r"), Loader=yaml.FullLoader)

        return few_shot_examples[self.dataset_name]


"""
The following are help functions for retrieving adversarial prompts from the markdown file.
"""
def split_markdown_by_title(markdown_file):
    with open(markdown_file, 'r', encoding='utf-8') as f:
        content = f.read()

    re_str = "# cola|# mnli|# mrpc|# qnli|# qqp|# rte|# sst2|# wnli|# mmlu|# squad_v2|# iwslt|# un_multi|# math"
    
    datasets = ["# cola", "# mnli", "# mrpc", "# qnli", "# qqp", "# rte", "# sst2", "# wnli",
                "# mmlu", "# squad_v2", "# iwslt", "# un_multi", "# math"]
    
    primary_sections = re.split(re_str, content)[1:]
    assert len(primary_sections) == len(datasets)

    all_sections_dict = {}

    for dataset, primary_section in zip(datasets, primary_sections):
        re_str = "## "
        results = re.split(re_str, primary_section)
        keywords = ["10 prompts", "bertattack", "checklist", "deepwordbug", "stresstest",
                    "textfooler", "textbugger", "translation"]

        secondary_sections_dict = {}
        for res in results:
            for keyword in keywords:
                if keyword in res.lower():
                    secondary_sections_dict[keyword] = res
                    break

        all_sections_dict[dataset] = secondary_sections_dict

    return all_sections_dict


def list_files(directory):
    files = [os.path.join(directory, d) for d in os.listdir(directory) if not os.path.isdir(os.path.join(directory, d))]
    return files


def convert_model_name(attack):
    attack_name = {
        "google/flan-t5-large": "t5",
        "google/flan-ul2": "ul2",
        "vicuna-13b": "vicuna",
        "llama2-13b-chat": "llama2",
        "chatgpt": "chatgpt",
    }
    return attack_name[attack]


def convert_dataset_name(dataset):
    return "# " + dataset


def retrieve(model_name, dataset_name, attack_name, prompt_type):
    model_name = convert_model_name(model_name)
    dataset_name = convert_dataset_name(dataset_name)

    if "zero" in prompt_type:
        shot = "zeroshot"
    else:
        shot = "fewshot"
    
    if "task" in prompt_type:
        prompt_type = "task"
    else:
        prompt_type = "role"
    
    directory_path = "./adv_prompts"
    md_dir = os.path.join(directory_path, model_name + "_" + shot + ".md")
    sections_dict = split_markdown_by_title(md_dir)
    results = {}
    for cur_dataset in sections_dict.keys():
        if cur_dataset == dataset_name:
            dataset_dict = sections_dict[cur_dataset]
            best_acc = 0
            best_prompt = ""
            for cur_attack in dataset_dict.keys():
                if cur_attack == "10 prompts":
                    prompts_dict = dataset_dict[cur_attack].split("\n")
                    num = 0
                    for prompt_summary in prompts_dict:
                        if "Acc: " not in prompt_summary:
                            continue
                        else:
                            import re
                            num += 1
                            match = re.search(r'Acc: (\d+\.\d+)%', prompt_summary)
                            if match:
                                number = float(match.group(1))
                                if number > best_acc:
                                    best_acc = number
                                    best_prompt = prompt_summary.split("prompt: ")[1]

            for cur_attack in dataset_dict.keys():
                
                if cur_attack == attack_name:

                    if attack_name == "translation":
                        prompts_dict = dataset_dict[attack_name].split("\n")
            
                        for prompt_summary in prompts_dict:
                            if "acc: " not in prompt_summary:
                                continue
                            
                            prompt = prompt_summary.split("prompt: ")[1]
                            
                            import re
                        
                            match_atk = re.search(r'acc: (\d+\.\d+)%', prompt_summary)
                            number_atk = float(match_atk.group(1))
                            results[prompt] = number_atk
  
                        sorted_results = sorted(results.items(), key=lambda item: item[1])[:6]

                        returned_results = []
                        for result in sorted_results:
                            returned_results.append({"origin prompt": best_prompt, "origin acc": best_acc, "attack prompt": result[0], "attack acc": result[1]})
      
                        return returned_results

                    elif attack_name in ["bertattack", "checklist", "deepwordbug", "stresstest", "textfooler", "textbugger"]:

                        prompts_dict = dataset_dict[attack_name].split("Original prompt: ")
                        num = 0
                        
                        returned_results = []
                        for prompt_summary in prompts_dict:
                            if "Attacked prompt: " not in prompt_summary:
                                continue

                            origin_prompt = prompt_summary.split("\n")[0]
                            attack_prompt = prompt_summary.split("Attacked prompt: ")[1].split("Original acc: ")[0]
                            attack_prompt = bytes(attack_prompt[2:-1], "utf-8").decode("unicode_escape").encode("latin1").decode("utf-8")

                            print(origin_prompt)
                            print(attack_prompt)

                            num += 1
                            import re
                            match_origin = re.search(r'Original acc: (\d+\.\d+)%', prompt_summary)
                            match_atk = re.search(r'attacked acc: (\d+\.\d+)%', prompt_summary)
                            if match_origin and match_atk:
                                if prompt_type == "task":
                                    if num > 3:
                                        break
                                else:
                                    if num < 3:
                                        continue
                                number_origin = float(match_origin.group(1))
                                number_atk = float(match_atk.group(1))
                                returned_results.append({"origin prompt": origin_prompt, "origin acc": number_origin, "attack prompt": attack_prompt, "attack acc": number_atk})
                        
                        return returned_results

    


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""
This file contains the prompt sets for the zeroshot role-oriented prompts.
"""

ROLE_ORIENTED_PROMPTS = {
    'valid_parentheses': [
        "As a syntax validator, assess the given sequence of brackets and determine whether it conforms to proper bracket rules. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",
        "In your role as an expression evaluator, analyze the provided arrangement of parentheses and ascertain its validity. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",
        "You are a bracket expert. Examine the sequence of brackets given and decide if it follows correct syntax rules. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",
        "As a parenthesis specialist, review the arrangement of brackets provided and determine whether it is a valid sequence. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",
        "In your capacity as a syntax verifier, analyze the string of brackets and identify if the order of parentheses is correct. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",
        "Investigate the validity of the given bracket sequence, ensuring it adheres to appropriate rules for a valid arrangement. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",
        "As a sequence assessor, evaluate the arrangement of brackets provided and ascertain whether it represents a valid expression. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",
        "You are a bracket authority. Examine the provided sequence of brackets and determine if it adheres to valid syntax conventions. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",
        "In your role as a syntax checker, assess the bracket arrangement and conclude whether it conforms to the requirements for validity. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",
        "As a bracket sequence expert, analyze the given string of brackets and judge whether it displays proper syntax and organization. Respond Valid if the brakets are matched, Invalid otherwise.\nQuestion: {content}\nAnswer:",    
    ], 
    
    'bool_logic': [
        "In your role as a logical evaluator, determine the truth value of the following expression. Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
        "As a computer scientist, evaluate the given boolean expression's truth value. Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
        "You are a debugging expert. Analyze the expression, determine its truth value. Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
        "As a software engineer, assess the boolean logic, what is the truth value? Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
        "In your capacity as a systems engineer, evaluate the expression, answer its truth value. Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
        "You are a decision-making specialist. Determine the truth value of this bool expression. Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
        "In your role as a problem solver, what is the truth value of this boolean expression? Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
        "As a logical thinker, what is the truth value of the following expression? Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
        "You are a code reviewer. Determine if the boolean logic is True or False. Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
        "In your role as a critical analyst, provide the bool expression's truth value. Respond True if it is, False otherwise.\nQuestion: {content}\nAnswer:",
    ],
    
    'math': [
        "As a mathematics instructor, calculate the answer to the following problem related to {task}: \nQuestion: {content}\nAnswer:",
        "Pretend you are a math teacher and solve this mathematical question concerning {task}: \nQuestion: {content}\nAnswer:",
        "Imagine being a math mentor and determine the solution to the following problem with {task}: \nQuestion: {content}\nAnswer:",
        "In the position of a mathematics coach, figure out the answer to this math challenge involving {task}: \nQuestion: {content}\nAnswer:",
        "Act as a math expert and compute the result of the following mathematical task about {task}: \nQuestion: {content}\nAnswer:",
        "As a mathematics consultant, resolve the following problem associated with {task}: \nQuestion: {content}\nAnswer:",
        "Taking on the role of a math specialist, work out the solution for this mathematical question featuring {task}: \nQuestion: {content}\nAnswer:",
        "Embodying a mathematics advisor, find the answer to the following math problem with {task}: \nQuestion: {content}\nAnswer:",
        "As a math guru, evaluate the following mathematical task regarding {task}: \nQuestion: {content}\nAnswer:",
        "In the shoes of a mathematics master, obtain the result for this math question that includes {task}: \nQuestion: {content}\nAnswer:",
    ],
# I want you to act as a prompt generator for squad v2 dataset.  
# Here is an example : "Please provide the most accurate answer based on the context. If the answer cannot be found in the context, respond with 'unanswerable'.". " 
# Please generate 10 similar prompts. the prompt is used for MMLU (Measuring Massive Multitask Language Understanding) dataset.  
# For the prompts, please first add a quote " at the beginning and the end of each sentence, and then and a comma at the end.
    'iwslt': [
        "Acting as a professional translator, translate the provided sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "As a skilled linguist, please convert the following sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "Assuming the role of an expert language translator, translate the given text from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "In the capacity of a proficient translator, please change the following sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "As a language specialist, accurately translate the provided sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "Taking on the role of a certified translator, perform a translation of the given sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "While acting as an experienced language expert, translate the following text from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "As a qualified translator, convert the given sentence from its original {soruce_lang} language to the target language {target_lang}: \nQuestion: {content}\nAnswer:",
        "Assuming the responsibilities of a professional translator, translate the subsequent text passage from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "In the role of a language expert, perform a machine translation for the provided sentence, changing it from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
    ],

    'un_multi': [
        "Acting as a professional translator, translate the provided sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "As a skilled linguist, please convert the following sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "Assuming the role of an expert language translator, translate the given text from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "In the capacity of a proficient translator, please change the following sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "As a language specialist, accurately translate the provided sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "Taking on the role of a certified translator, perform a translation of the given sentence from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "While acting as an experienced language expert, translate the following text from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "As a qualified translator, convert the given sentence from its original {soruce_lang} language to the target language {target_lang}: \nQuestion: {content}\nAnswer:",
        "Assuming the responsibilities of a professional translator, translate the subsequent text passage from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
        "In the role of a language expert, perform a machine translation for the provided sentence, changing it from {soruce_lang} to {target_lang}: \nQuestion: {content}\nAnswer:",
    ],

    'squad_v2': [
        "As a well-informed specialist familiar with the context, provide an answer to the question. If the context doesn't contain an answer, reply with 'unanswerable'.\nQuestion: {content}\nAnswer:",
        "Drawing upon your expertise in the context, determine the most suitable answer. If an answer isn't available, state 'unanswerable'.\nQuestion: {content}\nAnswer:",
        "As a subject matter expert, extract the correct answer from the context. If an answer is not present, indicate 'unanswerable'.\nQuestion: {content}\nAnswer:",
        "Using your knowledge of the context, identify the best answer to the question. If the context doesn't provide an answer, write 'unanswerable'.\nQuestion: {content}\nAnswer:",
        "As an authority on the context, locate the most accurate answer. If the context doesn't contain the answer, mention 'unanswerable'.\nQuestion: {content}\nAnswer:",
        "Being well-versed in the context, please derive the most fitting answer. If there isn't an answer in the context, use 'unanswerable'.\nQuestion: {content}\nAnswer:",
        "As an expert with a deep understanding of the context, find the best answer. If the context doesn't include an answer, say 'unanswerable'.\nQuestion: {content}\nAnswer:",
        "Drawing on your expertise in the context, provide the most precise answer. If the answer is not in the context, respond with 'unanswerable'.\nQuestion: {content}\nAnswer:",
        "As a proficient expert in the given context, search for the most relevant answer. If the answer cannot be found, respond by saying 'unanswerable'.\nQuestion: {content}\nAnswer:",
        "With your extensive knowledge of the context, answer the question accurately. If the context doesn't contain the answer, reply with 'unanswerable'."
    ],

    'mmlu': [
        "As an expert in {task}, respond to the following multiple-choice question by selecting 'A', 'B', 'C', or 'D'.\nQuestion: {content}\nAnswer:",
        "Given your proficiency in {task}, please answer the subsequent multiple-choice question with 'A', 'B', 'C', or 'D'.\nQuestion: {content}\nAnswer:",
        "With your knowledge of {task}, tackle the following multiple-choice question by choosing 'A', 'B', 'C', or 'D'.\nQuestion: {content}\nAnswer:",
        "As someone well-versed in {task}, please address the multiple-choice question below by selecting 'A', 'B', 'C', or 'D'.\nQuestion: {content}\nAnswer:",
        "Utilizing your expertise in {task}, answer the following multiple-choice question by picking 'A', 'B', 'C', or 'D'.\nQuestion: {content}\nAnswer:",
        "As a knowledgeable individual in {task}, provide your response to the multiple-choice question by choosing 'A', 'B', 'C', or 'D'.\nQuestion: {content}\nAnswer:",
        "With your understanding of {task}, kindly answer the subsequent multiple-choice question by selecting 'A', 'B', 'C', or 'D'.\nQuestion: {content}\nAnswer:",
        "As a skilled person in the field of {task}, please respond to the multiple-choice question by choosing 'A', 'B', 'C', or 'D'.\nQuestion: {content}\nAnswer:",
        "Considering your familiarity with {task}, attend to the following multiple-choice question by picking 'A', 'B', 'C', or 'D'.\nQuestion: {content}\nAnswer:",
        "Drawing upon your mastery of {task}, please answer the multiple-choice question by selecting the correct option from 'A', 'B', 'C', or 'D'."
    ],

    'sst2': [
        "As a sentiment classifier, determine whether the following text is 'positive' or 'negative'. Please classify: \nQuestion: {content}\nAnswer:",
        "In the role of a sentiment analysis tool, respond with 'positive' or 'negative' to classify this statement: \nQuestion: {content}\nAnswer:",
        "Acting as a sentiment evaluator, identify if the given sentence is 'positive' or 'negative'. Classify: \nQuestion: {content}\nAnswer:",
        "As an emotion detector, determine if the provided passage conveys a 'positive' or 'negative' sentiment. Classify: \nQuestion: {content}\nAnswer:",
        "Working as a sentiment analyzer, please indicate if the following text is 'positive' or 'negative'. Classify: \nQuestion: {content}\nAnswer:",
        "In the capacity of a sentiment classifier, decide whether the given quote is 'positive' or 'negative'. Classify: \nQuestion: {content}\nAnswer:",
        "Taking on the role of an emotion classifier, specify if the provided phrase is 'positive' or 'negative'. Classify: \nQuestion: {content}\nAnswer:",
        "Functioning as a sentiment identification tool, assess if the following expression is 'positive' or 'negative'. Classify: \nQuestion: {content}\nAnswer:",
        "Serving as a sentiment evaluation model, determine if the given statement is 'positive' or 'negative'. Classify: \nQuestion: {content}\nAnswer:",
        "Emulating a sentiment classification system, indicate whether the provided text is 'positive' or 'negative'. Classify: \nQuestion: {content}\nAnswer:",
    ],

    # 56.34, 66.20, 61.97, 59.15, 59.15, 56.34, 64.79, 57.75, 64.79, 54.93
    'wnli': [
        "In your role as an entailment analysis tool, assess the relationship between the given sentences and classify it as 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "As an entailment identification system, examine the connection between the following sentences and respond with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "Functioning as an entailment evaluation tool, analyze the provided sentences and decide if their relationship is 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "Acting as an entailment detection instrument, determine if the given pair of sentences demonstrates entailment or not_entailment. Answer with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "As a tool for determining entailment relationships, review the two statements and categorize their connection as either 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "While performing entailment analysis, classify the relationship between the provided sentences as 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "In the capacity of an entailment assessment system, indicate if the link between the following sentences is 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "Working as an entailment classifier, identify whether the given pair of sentences displays entailment or not_entailment. Respond with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "As an instrument for entailment evaluation, consider the two sentences and determine if their relationship is 'entailment' or 'not_entailment'. Respond with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "In the role of a semantic relationship analyzer, examine the connection between the given sentences and decide if they exhibit entailment or not_entailment. Answer with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",   
    ],

    # 84.48, 84.12, 84.48, 84.48, 84.12, 84.84, 84.84, 83.03, 85.56, 82.31
    'rte': [
        "In your role as an entailment analysis tool, assess the relationship between the given sentences and classify it as 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "As an entailment identification system, examine the connection between the following sentences and respond with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "Functioning as an entailment evaluation tool, analyze the provided sentences and decide if their relationship is 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "Acting as an entailment detection instrument, determine if the given pair of sentences demonstrates entailment or not_entailment. Answer with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "As a tool for determining entailment relationships, review the two statements and categorize their connection as either 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "While performing entailment analysis, classify the relationship between the provided sentences as 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "In the capacity of an entailment assessment system, indicate if the link between the following sentences is 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "Working as an entailment classifier, identify whether the given pair of sentences displays entailment or not_entailment. Respond with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "As an instrument for entailment evaluation, consider the two sentences and determine if their relationship is 'entailment' or 'not_entailment'. Respond with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",
        "In the role of a semantic relationship analyzer, examine the connection between the given sentences and decide if they exhibit entailment or not_entailment. Answer with 'entailment' or 'not_entailment':\nQuestion: {content}\nAnswer:",       
    ],

    'mnli': [
        "In your role as an entailment analysis tool, assess the relationship between the given sentences and classify it as 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
        "As an entailment identification system, examine the connection between the following sentences and respond with 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
        "Functioning as an entailment evaluation tool, analyze the provided sentences and decide if their relationship is 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
        "Acting as an entailment detection instrument, determine if the given pair of sentences demonstrates entailment, neutral, or contradiction. Answer with 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
        "As a tool for determining entailment relationships, review the two statements and categorize their connection as either 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
        "While performing entailment analysis, classify the relationship between the provided sentences as 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
        "In the capacity of an entailment assessment system, indicate if the link between the following sentences is 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
        "Working as an entailment classifier, identify whether the given pair of sentences displays entailment, neutral, or contradiction. Respond with 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
        "As an instrument for entailment evaluation, consider the two sentences and determine if their relationship is 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
        "In the role of a semantic relationship analyzer, examine the connection between the given sentences and decide if they exhibit entailment, neutral, or contradiction. Answer with 'entailment', 'neutral', or 'contradiction':\nQuestion: {content}\nAnswer:",
    ],

    'cola': [
        "In your role as a grammar check tool, assess the following sentence and classify it as 'acceptable' if it is grammatically correct or 'unacceptable' if it is incorrect:\nQuestion: {content}\nAnswer:",
        "As a grammar identification system, examine the provided sentence and respond with 'acceptable' for grammatically correct sentences or 'unacceptable' for incorrect ones:\nQuestion: {content}\nAnswer:",
        "Functioning as a grammar evaluation tool, analyze the given sentence and decide if it is grammatically correct, responding with 'acceptable' or 'unacceptable':\nQuestion: {content}\nAnswer:",
        "Acting as a grammar detection instrument, determine if the provided sentence is grammatically sound, answering with 'acceptable' for correct grammar or 'unacceptable' for incorrect grammar:\nQuestion: {content}\nAnswer:",
        "As a tool for determining grammatical correctness, review the sentence and categorize its grammar as either 'acceptable' or 'unacceptable':\nQuestion: {content}\nAnswer:",
        "While performing grammar analysis, classify the grammar of the following sentence as 'acceptable' for correct grammar or 'unacceptable' for incorrect grammar:\nQuestion: {content}\nAnswer:",
        "In the capacity of a grammar assessment system, indicate if the structure of the provided sentence is grammatically correct, responding with 'acceptable' or 'unacceptable':\nQuestion: {content}\nAnswer:",
        "Working as a grammar classifier, identify whether the given sentence has correct grammar, and respond with 'acceptable' for correct sentences or 'unacceptable' for incorrect ones:\nQuestion: {content}\nAnswer:",
        "As an instrument for grammar evaluation, consider the sentence and determine if its grammar is correct, responding with 'acceptable' for correct grammar or 'unacceptable' for incorrect grammar:\nQuestion: {content}\nAnswer:",
        "In the role of a syntax analyzer, examine the grammar of the provided sentence and decide if it is correct, answering with 'acceptable' for grammatically correct sentences or 'unacceptable' for incorrect ones:\nQuestion: {content}\nAnswer:",
    ],

    'qqp': [
        "In your role as a question comparison tool, assess the following pair of questions and classify them as 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "As a question equivalence detection system, examine the provided questions and respond with 'equivalent' if they are the same in meaning, or 'not_equivalent' if they are different. \nQuestion: {content}\nAnswer:",
        "Functioning as a question similarity evaluation tool, analyze the given questions and decide if they share the same meaning, responding with 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "Acting as a question equivalence instrument, determine if the provided questions are equivalent in meaning, answering with 'equivalent' for similar questions or 'not_equivalent' for dissimilar ones. \nQuestion: {content}\nAnswer:",
        "As a tool for determining question equivalence, review the questions and categorize their similarity as either 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "While performing question comparison analysis, classify the similarity of the following questions as 'equivalent' for equivalent questions or 'not_equivalent' for different questions. \nQuestion: {content}\nAnswer:",
        "In the capacity of a question assessment system, indicate if the meaning of the provided questions is the same, responding with 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "Working as a question classifier, identify whether the given questions share the same meaning, and respond with 'equivalent' for equivalent questions or 'not_equivalent' for different ones. \nQuestion: {content}\nAnswer:",
        "As an instrument for question comparison evaluation, consider the questions and determine if their meaning is the same, responding with 'equivalent' for similar questions or 'not_equivalent' for different questions. \nQuestion: {content}\nAnswer:",
        "In the role of a question similarity analyzer, examine the meaning of the provided questions and decide if they are equivalent, answering with 'equivalent' for equivalent questions or 'not_equivalent' for different questions. \nQuestion: {content}\nAnswer:",
    ],

    # 86.95, 88.65, 88.85, 87.90, 83.10, 74.45, 88.55, 88.65, 88.85, 83.80
    'qnli': [
        "As a language expert, assess if the given context entails the answer to the question and respond with 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",
        "In your role as a semantic evaluator, determine if the provided context justifies the answer to the question and answer with 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",
        "As a textual analyst, examine if the given context logically implies the answer to the question and indicate your decision with 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",
        "As a semantic researcher, evaluate whether the provided context supports the answer to the question and choose 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",
        "In the capacity of a language specialist, decide if the context presented contains enough information to infer the answer to the question and respond with 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",
        "As a textual inference expert, analyze if the answer to the question can be deduced from the provided context and select 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",
        "In your role as a linguistic investigator, determine if the context given entails the answer to the question and provide your conclusion with 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",
        "As a semantic interpreter, assess whether the provided context supports the answer to the given question and answer with 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",
        "In the capacity of a language evaluator, examine if the given context justifies the answer to the question and indicate your assessment with 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",
        "As a linguistic consultant, decide if the answer to the question is logically supported by the provided context and respond with 'entailment' or 'not_entailment'. \nQuestion: {content}\nAnswer:",    
    ],


    # 82.60, 77.94, 80.39, 81.13, 80.64, 75.74, 81.62, 81.13, 79.66, 82.60
    'mrpc': [
        "As a semantic comparison expert, evaluate the given pair of sentences and determine if they are 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "In your capacity as a language analyst, assess the following sentences and classify their similarity as 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "As a sentence similarity evaluator, analyze the provided sentences and indicate if their meanings are 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "In the role of a textual comparison specialist, examine the given sentences and decide if they share the same meaning, responding with 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "As a linguistic comparator, review the following pair of sentences and determine their semantic equivalence by choosing 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "In your capacity as a semantic assessment tool, evaluate the provided sentences and classify their meanings as 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "As a language comparison expert, examine the given pair of sentences and decide if their meanings align, answering with 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "In the role of a sentence comparison analyst, assess the provided sentences and indicate if they convey the same meaning by selecting 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "As a textual similarity evaluator, analyze the following pair of sentences and determine if they are semantically 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:",
        "In your capacity as a semantic comparison tool, examine the given sentences and decide if their meanings are identical, responding with 'equivalent' or 'not_equivalent'. \nQuestion: {content}\nAnswer:", 
    ],
}


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import os
import re
import random
import requests
import json
from PIL import Image as PILImage
from tqdm import tqdm

from promptbench.config import *
from datasets import load_dataset


def shuffleDict(d):
    keys = list(d.keys())
    random.shuffle(keys)
    [(key, d[key]) for key in keys]
    random.shuffle(keys)
    [(key, d[key]) for key in keys]
    random.shuffle(keys)
    keys = [(key, d[key]) for key in keys]
    #keys = d(keys)
    return dict(keys)

class Dataset(object):
    def __init__(self, dataset_name):
        self.data = []
        self.dataset_name = dataset_name

        # Get the parent directory
        cur_dir = os.path.dirname(os.path.abspath(__file__))
        self.data_dir = os.path.join(os.path.dirname(cur_dir), 'data')

        if not os.path.exists(self.data_dir):
            os.mkdir(self.data_dir)
        
        # check if the dataset exists, if not, download it
        if dataset_name == "gsm8k":
            self.filepath = os.path.join(self.data_dir, f"{dataset_name}.jsonl")
        else:
            self.filepath = os.path.join(self.data_dir, f"{dataset_name}.json")
            
        if not os.path.exists(self.filepath):
            if dataset_name == "gsm8k":
                url = f'https://wjdcloud.blob.core.windows.net/dataset/promptbench/dataset/{dataset_name}.jsonl'
            else:
                url = f'https://wjdcloud.blob.core.windows.net/dataset/promptbench/dataset/{dataset_name}.json'
            print(f"Downloading {dataset_name} dataset...")
            response = requests.get(url)
            with open(self.filepath, 'wb') as f:
                f.write(response.content)

    def __len__(self):
        assert len(self.data) > 0, "Empty dataset. Please load data first."
        return len(self.data)

    def __getitem__(self, idx):
        assert len(self.data) > 0, "Empty dataset. Please load data first."
        return self.data[idx]
    
    def extract_answer(self, output): 
        return output

    def save_images_to_local(self, dataset, split, key_list):
        # Get the parent directory
        cur_dir = os.path.dirname(os.path.abspath(__file__))

        self.data_dir = os.path.join(os.path.dirname(cur_dir), 'data')
        if not os.path.exists(self.data_dir):
            os.mkdir(self.data_dir)

        self.dataset_dir = os.path.join(self.data_dir, dataset)
        if not os.path.exists(self.dataset_dir):
            os.mkdir(self.dataset_dir)
        
        self.split_dir = os.path.join(self.dataset_dir, split)
        if not os.path.exists(self.split_dir):
            os.mkdir(self.split_dir)

        self.filepath = os.path.join(self.dataset_dir, f"{split}.json")
        if not os.path.exists(self.filepath):
            local_path_dict = {}
            print('Saveing images to local path: ', self.split_dir)
            for idx in tqdm(range(len(self.data))):
                local_path_dict[str(idx)] = []
                for key in key_list:
                    if self.data[idx][key] is not None:
                        image_dir = os.path.join(self.split_dir, f'{idx}_{key}.png')
                        self.data[idx][key].save(image_dir)
                        local_path_dict[str(idx)].append(image_dir)
            print('Saving file: ', self.filepath)
            with open(self.filepath, 'w') as f:
                json.dump(local_path_dict, f)
        else:
            print('Images already saved to local, loading file: ', self.filepath)
            with open(self.filepath, 'r') as f:
                local_path_dict = json.load(f)

        self.local_path_dict = local_path_dict

class BoolLogic(Dataset):
    """
    BoolLogic is a dataset class (obtained from BigBench) representing boolean logic expressions. Each entry in the dataset consists of a boolean logic expression and its corresponding label (either 'true' or 'false'). The dataset is read from a JSON file containing questions and answers.
    
    https://github.com/google/BIG-bench/tree/main/bigbench/benchmark_tasks/boolean_expressions

    Example data format:
    [{'content': 'not ( not False and True ) or not True is ', 'label': 'false'}, ...]
    """
    def __init__(self):
        super().__init__("bool_logic")
        with open(self.filepath, 'r') as f:
            data = json.load(f)
        
        self.data = [{"content": d["question"], "label": "true" if d["answer"] else "false"} for d in data]


class ValidParentheses(Dataset):
    """
    ValidParentheses is a dataset class (obtained from BigBench) for validating parentheses in strings. It checks if the given string of parentheses is valid or invalid. The dataset is initialized by reading from a JSON file, with each entry containing a string of parentheses and a label ('valid' or 'invalid').

    https://github.com/google/BIG-bench/tree/main/bigbench/benchmark_tasks/cs_algorithms

    Example data format:
    [{'content': '( ] } (', 'label': 'invalid'}, ...]
    """
    def __init__(self):
        super().__init__("valid_parentheses")

        with open(self.filepath, 'r') as f:
            data = json.load(f)["examples"][:100]
        for d in data:
            self.data.append(
                {"content": d["input"], "label": "valid" if d["target_scores"]["Valid"] == 1 else "invalid"})


class Math(Dataset):
    """
    Math is a dataset class that loads mathematical questions and answers from the Hugging Face datasets (math_dataset test set). This dataset covers various types of math questions, such as algebra, calculus, and arithmetic. It is initialized with a specific type of math question.

    Reference: https://huggingface.co/datasets/math_dataset/

    Example data format:
    [{'question': "Solve -282*d + 929 - 178 = -1223 for d.\\n'", 'answer': "b'7\\n'", 'task': 'algebra__linear_1d'}, ...]
    """
    def __init__(self, task) -> None:

        self.data = []

        MATH_QUESTION_TYPES = [
            'algebra__linear_1d',
            'algebra__linear_1d_composed',
            'algebra__linear_2d',
            'algebra__linear_2d_composed',
            'algebra__polynomial_roots',
            'algebra__polynomial_roots_composed',
            'algebra__sequence_next_term',
            'algebra__sequence_nth_term',
            'arithmetic__add_or_sub',
            'arithmetic__add_or_sub_in_base',
            'arithmetic__add_sub_multiple',
            'arithmetic__div',
            'arithmetic__mixed',
            'arithmetic__mul',
            'arithmetic__mul_div_multiple',
            'arithmetic__nearest_integer_root',
            'arithmetic__simplify_surd',

            'calculus__differentiate',
            'calculus__differentiate_composed',

            'comparison__closest',
            'comparison__closest_composed',
            'comparison__kth_biggest',
            'comparison__kth_biggest_composed',
            'comparison__pair',
            'comparison__pair_composed',
            'comparison__sort',
            'comparison__sort_composed',

            'measurement__conversion',
            'measurement__time',

            'numbers__base_conversion',
            'numbers__div_remainder',
            'numbers__div_remainder_composed',
            'numbers__gcd',
            'numbers__gcd_composed',
            'numbers__is_factor',
            'numbers__is_factor_composed',
            'numbers__is_prime',
            'numbers__is_prime_composed',
            'numbers__lcm',
            'numbers__lcm_composed',
            'numbers__list_prime_factors',
            'numbers__list_prime_factors_composed',
            'numbers__place_value',
            'numbers__place_value_composed',
            'numbers__round_number',
            'numbers__round_number_composed',

            'polynomials__add',
            'polynomials__coefficient_named',
            'polynomials__collect',
            'polynomials__compose',
            'polynomials__evaluate',
            'polynomials__evaluate_composed',
            'polynomials__expand',
            'polynomials__simplify_power',
            'probability__swr_p_level_set',
            'probability__swr_p_sequence',
        ]
        
        assert task in MATH_QUESTION_TYPES, f"Task {task} is not supported in math dataset. Please choose from {MATH_QUESTION_TYPES}"
        data = load_dataset("math_dataset", task)["test"]
        for d in data:
            d["task"] = task
            self.data.append(d)


class UnMulti(Dataset):
    """
    UnMulti is a dataset class for multilingual translation tasks. It includes translations between multiple language pairs. The dataset is partially loaded from a JSON file due to its large size.

    Example data format:
    [{'source': '4 - العميد بحري مرتضى سفاري، قائد القوات البحرية', 'target': 'Konteradmiral Morteza Safari (Kommandeur der Marine des Korps der Iranischen Revolutionsgarden)', 'soruce_lang': 'Arabic', 'target_lang': 'German'}, ...]
    """
    def __init__(self):
        # un_multi dataset only have train set and is large, so we selected partial dataset.
        super().__init__("un_multi")

        with open(self.filepath, 'r') as f:
            data = json.load(f)
        self.data = []

        LANGUAGES = {
            'ar': 'Arabic',
            'de': 'German',
            'en': 'English',
            'es': 'Spanish',
            'fr': 'French',
            'ru': 'Russian',
            'zh': 'Chinese',
            'it': 'Italian',
            'nl': 'Dutch',
            'ro': 'Romanian',
            'ja': 'Japanese',
            'ko': 'Korean',
        }

        for task_i in data.keys():
            source, target = task_i.split('-')
            for d in data[task_i]:
                self.data.append({
                    'source': d[source],
                    'target': d[target],
                    'soruce_lang': LANGUAGES[source],
                    'target_lang': LANGUAGES[target],
                })


class IWSLT(Dataset):
    """
    IWSLT is a dataset class for the International Workshop on Spoken Language Translation. It includes pairs of sentences in different languages, intended for translation tasks. The dataset is loaded from Hugging Face datasets and supports multiple language pairs.

    Reference: https://huggingface.co/datasets/iwslt

    Example data format:
    [{'source': 'قبل عدة سنوات، هنا في تيد، قدّم بيتر سكيلمان منافسة تصميم تسمى منافسة حلوى المارش مالو.', 'target': 'Several years ago here at TED, Peter Skillman introduced a design challenge called the marshmallow challenge.', 'soruce_lang': 'Arabic', 'target_lang': 'English'}, ...]
    """
    def __init__(self, supported_languages):
        # IWSLT now is loaded from huggingface datasets.
        # https://huggingface.co/datasets/iwslt

        self.data = []

        LANGUAGES = {
            'ar': 'Arabic',
            'de': 'German',
            'en': 'English',
            'es': 'Spanish',
            'fr': 'French',
            'ru': 'Russian',
            'zh': 'Chinese',
            'it': 'Italian',
            'nl': 'Dutch',
            'ro': 'Romanian',
            'ja': 'Japanese',
            'ko': 'Korean',
        }

        for task_i in supported_languages:
            source, target = task_i.split('-')
            data = load_dataset("iwslt2017", "iwslt2017-"+task_i)["test"]
            for d in data:
                d = d["translation"]
                self.data.append({
                    'source': d[source],
                    'target': d[target],
                    'soruce_lang': LANGUAGES[source],
                    'target_lang': LANGUAGES[target],
                })


class SQUAD_V2(Dataset):
    """
    SQUAD_V2 is a dataset class for the Stanford Question Answering Dataset (SQuAD) version 2, which involves question-answering tasks.
    SQUAD_V2 dataset is loaded from huggingface datasets.
    Reference: https://huggingface.co/datasets/squad_v2

    Example data format:
    [{'id': '56ddde6b9a695914005b9628', 'title': 'Normans', 'context': 'The Normans (Norman: Nourmands; French: Normands; Latin: Normanni) were the people who in the 10th and 11th centuries gave their name to Normandy, a region in France. They were descended from Norse ("Norman" comes from "Norseman") raiders and pirates from Denmark, Iceland and Norway who, under their leader Rollo, agreed to swear fealty to King Charles III of West Francia. Through generations of assimilation and mixing with the native Frankish and Roman-Gaulish populations, their descendants would gradually merge with the Carolingian-based cultures of West Francia. The distinct cultural and ethnic identity of the Normans emerged initially in the first half of the 10th century, and it continued to evolve over the succeeding centuries.', 'question': 'In what country is Normandy located?', 'answers': {'text': ['France', 'France', 'France', 'France'], 'answer_start': [159, 159, 159, 159]}}, ...]
    """
    def __init__(self):
        # 
        # 
        self.data = []
        data = load_dataset("squad_v2")["validation"]
        for d in data:
            self.data.append(d)


class MMLU(Dataset):
    """
    MMLU is a dataset class for the Multimodal Multi-Task Learning Understanding dataset, covering various educational and professional fields.
    MMLU dataset is loaded from huggingface datasets: lukaemon/mmlu (test set).
    
    Reference: https://huggingface.co/datasets/cais/mmlu/viewer/abstract_algebra/test

    Example data format:
    [{'input': "This question refers to the following information.\nRead the the following quotation to answer questions.\nThe various modes of worship which prevailed in the Roman world were all considered by the people as equally true; by the philosopher as equally false; and by the magistrate as equally useful.\nEdward Gibbon, The Decline and Fall of the Roman Empire, 1776–1788\nGibbon's interpretation of the state of religious worship in ancient Rome could be summarized as", 'A': "In ancient Rome, religious worship was decentralized and tended to vary with one's social position.", 'B': 'In ancient Rome, religious worship was the source of much social tension and turmoil.', 'C': 'In ancient Rome, religious worship was homogeneous and highly centralized.', 'D': 'In ancient Rome, religious worship was revolutionized by the introduction of Christianity.', 'target': 'A', 'task': 'high_school_european_history'}, ...]
    """
    def __init__(self):       
        self.data = []
        self.tasks = ['high_school_european_history', 'business_ethics', 'clinical_knowledge', 'medical_genetics',
                    'high_school_us_history', 'high_school_physics', 'high_school_world_history', 'virology',
                    'high_school_microeconomics', 'econometrics', 'college_computer_science', 'high_school_biology',
                    'abstract_algebra', 'professional_accounting', 'philosophy', 'professional_medicine', 'nutrition',
                    'global_facts', 'machine_learning', 'security_studies', 'public_relations', 'professional_psychology',
                    'prehistory', 'anatomy', 'human_sexuality', 'college_medicine', 'high_school_government_and_politics',
                    'college_chemistry', 'logical_fallacies', 'high_school_geography', 'elementary_mathematics', 'human_aging',
                    'college_mathematics', 'high_school_psychology', 'formal_logic', 'high_school_statistics', 'international_law',
                    'high_school_mathematics', 'high_school_computer_science', 'conceptual_physics', 'miscellaneous', 'high_school_chemistry',
                    'marketing', 'professional_law', 'management', 'college_physics', 'jurisprudence', 'world_religions', 'sociology',
                    'us_foreign_policy', 'high_school_macroeconomics', 'computer_security', 'moral_scenarios', 'moral_disputes',
                    'electrical_engineering', 'astronomy', 'college_biology']

        for task in self.tasks:
            data = load_dataset("cais/mmlu", task)["test"]
            for d in data:
                d["task"] = task
                self.data.append(d)

            
class GLUE(Dataset):
    """
    GLUE class is a dataset class for the General Language Understanding Evaluation benchmark, supporting multiple natural language understanding tasks.

    Examples:
    [{'content': "it 's a charming and often affecting journey . ", 'label': 1}, {'content': 'unflinchingly bleak and desperate ', 'label': 0}, ...]
    """
    def __init__(self, task):
        self.data = []
        self.supported_tasks = ["sst2", "cola", "qqp", "mnli",
                                "mnli_matched", "mnli_mismatched", "qnli", "wnli", "rte", "mrpc"]
        assert task in self.supported_tasks

        self.task = task

        if self.task == "mnli":
            from datasets import concatenate_datasets
            matched = load_dataset('glue', 'mnli')["validation_matched"]
            mismatched = load_dataset("glue", "mnli")["validation_mismatched"]
            data = concatenate_datasets([matched, mismatched])
        else:
            data = load_dataset("glue", task)["validation"]
        
        for d in data:
            if task == "sst2" or task == "cola":
                content = d['sentence']
            elif task == 'qqp':
                content = 'Question 1: ' + \
                    d['question1'] + ' Question 2: ' + \
                    d['question2']
            elif task == 'mnli' or task == 'mnli_matched' or task == 'mnli_mismatched':
                content = 'Premise: ' + \
                    d['premise'] + ' Hypothesis: ' + \
                    d['hypothesis']
            elif task == 'qnli':
                content = 'Question: ' + \
                    d['question'] + ' Context: ' + \
                    d['sentence']
            elif task == 'rte' or task == 'mrpc' or task == "wnli":
                content = 'Sentence 1: ' + \
                    d['sentence1'] + ' Sentence 2: ' + \
                    d['sentence2']
            else:
                raise NotImplementedError

            self.data.append({"content": content, "label": d['label']})            


class GSM8K(Dataset):
    """
    GSM8K is a dataset class that loads mathematical questions and answers from the Hugging Face datasets (main test set). The dataset is a collection of 8.5K high-quality, linguistically diverse grade school math word problems.
    GSM8K dataset is loaded from huggingface datasets: /gsm8k (test set).
    
    Reference: https://huggingface.co/datasets/gsm8k/viewer/main/test
    
    Example data format:
    [{'question': "A robe takes 2 bolts of blue fiber and half that much white fiber. How many bolts in total does it take?", 'answer': "It takes 2/2=<<2/2=1>>1 bolt of white fiber So the total amount of fabric is 2+1=<<2+1=3>>3 bolts of fabric #### 3"}, ...]
    """
    def __init__(self):
        # gsm8k dataset now is loaded from huggingface datasets: /gsm8k (test set).
        # https://huggingface.co/datasets/gsm8k
        
        data = load_dataset("gsm8k", "main")["test"]
        self.data = []
        for d in data:
            content = d["question"].strip()
            label = d["answer"].split("#### ")[-1]
            self.data.append({"content": content, "label": label})
    
    def extract_answer(self, output):
        answer = output.replace(",", "")
        answer = [s for s in re.findall(r'-?\d+\.?\d*', answer)]
        answer = answer[0] if len(answer) > 0 else ""
        return answer


class BigBench(Dataset):
    """
    BigBench is a dataset class that loads questions and answers from the BigBench benchmark (sub-dataset: date and object tracking). It includes various types of questions, such as date understanding, object tracking. The dataset is loaded from a JSON file containing questions and answers.
    
    date: https://github.com/google/BIG-bench/blob/main/bigbench/benchmark_tasks/date_understanding/task.json
    object_tracking: https://github.com/google/BIG-bench/blob/main/bigbench/benchmark_tasks/logical_deduction/three_objects/task.json
    
    Example data format:
    [
        {
            "input": "On a shelf, there are three books: a black book, an orange book, and a blue book. The blue book is to the right of the orange book. The orange book is to the right of the black book.",
            "target_scores": {
                "The black book is the leftmost.": 1,
                "The orange book is the leftmost.": 0,
                "The blue book is the leftmost.": 0
            }
        },
    """
    def __init__(self, dataset_name):
        super().__init__(dataset_name)
        
        self.data = []
        with open(self.filepath, 'r') as f:
            json_data = json.load(f)
            json_data = json_data["examples"]
            if self.dataset_name == "bigbench_date":
                choice_index = ['A','B','C','D','E','F']
            elif self.dataset_name == "bigbench_object_tracking":
                choice_index = ['A','B','C']
            else:
                raise ValueError("dataset is not properly defined ...")
            for line in json_data:
                raw_q = line["input"].strip()
                if self.dataset_name == "bigbench_date":
                    choice = "Answer Choices:"
                    # Randomly shuffle the answer choice dictionary because the original answer is always A ...
                    choice_dic = shuffleDict(line["target_scores"])
                elif self.dataset_name == "bigbench_object_tracking":
                    choice = "\nWhich choice is true ? Answer Choices:"
                    choice_dic = line["target_scores"]
                else:
                    raise ValueError("dataset is not properly defined ...")
                for i, key_value in enumerate(choice_dic.items()):
                    key, value = key_value
                    choice += " ("
                    choice += choice_index[i]
                    choice += ") "
                    choice += key
                    if value == 1:
                        a = choice_index[i]
                        #a = key
                q = raw_q + " " + choice
                self.data.append({"content": q, "label": a})
    
    def extract_answer(self, output): 
        if self.dataset_name == "bigbench_date":
            answer = re.findall(r'A|B|C|D|E|F', output)
        elif self.dataset_name == "bigbench_object_tracking":
            answer = re.findall(r'A|B|C', output)
        
        answer = answer[0] if len(answer) > 0 else ""
        
        # print(answer)
        return answer

             
class CSQA(Dataset):
    """
    CSQA is a dataset class that loads questions and answers from the CommonsenseQA dataset. The dataset is a challenging commonsense question-answering dataset. It comprises 12,247 questions with 5 multiple-choice answers each. 
    CSQA dataset now is loaded from huggingface datasets: /commonsense_qa (val set).
    
    Reference: https://huggingface.co/datasets/commonsense_qa/viewer/default/validation
    
    Example data format:
    ['id': "1afa02df02c908a558b4036e80242fac", 'question': "A revolving door is convenient for two direction travel, but it also serves as a security measure at a what?", 'question_concept': "revolving door", 'choices': { "label": [ "A", "B", "C", "D", "E" ], "text": [ "bank", "library", "department store", "mall", "new york" ] }, 'answerKey': "A" ]
    
    """
    def __init__(self):
        data = load_dataset("commonsense_qa")["validation"]
        
        self.data = []
        choice_index = ['A','B','C','D','E']
        for d in data:
            raw_q = d["question"].strip()
            choice = "\nAnswer Choices:"
            choice_list = d["choices"]["text"]
            for i, c in enumerate(choice_list):
                choice += " ("
                choice += choice_index[i]
                choice += ") "
                choice += c
            q = raw_q + " " + choice
            a = d["answerKey"]
            self.data.append({"content": q, "label": a})
                
    def extract_answer(self, output): 
        answer = re.findall(r'A|B|C|D|E', output)
        answer = answer[0] if len(answer) > 0 else ""
        
        # print(answer)
        return answer
    

class LastLetterConcat(Dataset):
    """
    LastLetterConcat is a dataset class that loads questions and answers from the Last Letter Concat dataset. The dataset is initialized by reading from a JSON file, with each entry containing a query and a answer.
    
    Reference: https://arxiv.org/pdf/2201.11903.pdf (page 8)
    
    Example data format:
    [{"question": "Take the last letters of each words in \"Whitney Erika Tj Benito\" and concatenate them.", "answer": "yajo"}, ...]
    """
    def __init__(self):
        super().__init__("last_letter_concat")
        
        self.data = []
        with open(self.filepath, 'r') as f:
            json_data = json.load(f)
            for line in json_data:
                q = line["question"].strip()
                a = line["answer"]
                self.data.append({"content": q, "label": a})
    

class NumerSense(Dataset):
    """
    NumerSense is a dataset class that loads questions and answers from the NumerSense dataset which is a unique numerical commonsense reasoning probing task. The dataset is initialized by reading from a JSON file, with each entry containing a query and a answer (English number).
    
    https://github.com/INK-USC/NumerSense/blob/main/data/validation.masked.tsv
    
    Example data format:
    [{"query": "you may take the subway back and forth to work <mask> days a week.", "answer": "five"}, ...]
    
    """
    def __init__(self):
        super().__init__("numersense")
        
        self.data = []
        with open(self.filepath, 'r') as f:
            json_data = json.load(f)
            for line in json_data:
                q = line["query"].strip()
                a = line["answer"]
                self.data.append({"content": q, "label": a})
                

class QASC(Dataset):
    """
    QASC is a dataset class that loads questions and answers from the Question Answering via QASC dataset. QASC is a question-answering dataset with a focus on sentence composition. It consists of 9,980 8-way multiple-choice questions about grade school science (8,134 train, 926 dev, 920 test), and comes with a corpus of 17M sentences.
    QASC dataset now is loaded from huggingface datasets: /qasc (val set).
    
    Reference: https://huggingface.co/datasets/qasc/viewer/default/validation
    
    Example data format: 
    ['id': "3NGI5ARFTT4HNGVWXAMLNBMFA0U1PG", 'question': "Climate is generally described in terms of what?", 'choices': { "text": [ "sand", "occurs over a wide range", "forests", "Global warming", "rapid changes occur", "local weather conditions", "measure of motion", "city life" ], "label": [ "A", "B", "C", "D", "E", "F", "G", "H" ] }, 'answerKey': "F", 'fact1': "Climate is generally described in terms of temperature and moisture.", 'fact2': "Fire behavior is driven by local weather conditions such as winds, temperature and moisture.", 'combinedfact':"Climate is generally described in terms of local weather conditions", 'formatted_question': "Climate is generally described in terms of what? (A) sand (B) occurs over a wide range (C) forests (D) Global warming (E) rapid changes occur (F) local weather conditions (G) measure of motion (H) city life"]
    
    """
    
    def __init__(self):
        data = load_dataset("qasc")["validation"]
        
        self.data = []
        choice_index = ['A','B','C','D','E','F','G','H']
        for d in data:
            raw_q = d["question"].strip()
            choice = "\nAnswer Choices:"
            choice_list = d["choices"]["text"]
            for i, c in enumerate(choice_list):
                choice += " ("
                choice += choice_index[i]
                choice += ") "
                choice += c
            q = raw_q + " " + choice
            a = d["answerKey"]
            self.data.append({"content": q, "label": a})
                
    def extract_answer(self, output): 
        answer = re.findall(r'A|B|C|D|E|F|G|H', output)
        answer = answer[0] if len(answer) > 0 else ""
        
        # print(answer)
        return answer

class BBH(Dataset):
    """
    BBH is a dataset class for the BigBench Hard dataset. This dataset is loaded from huggingface datasets: lukaemon/bbh (test set).
    
    Reference: 
    https://huggingface.co/datasets/lukaemon/bbh
    https://github.com/suzgunmirac/BIG-Bench-Hard

    Example data format:
    {'input': 'not ( True ) and ( True ) is', 'target': 'False', 'task': 'boolean_expressions'}
    """
    def __init__(self):       
        self.data = []
        self.tasks = ['boolean_expressions', 'causal_judgement', 'date_understanding', 'disambiguation_qa', 
                      'dyck_languages', 'formal_fallacies', 'geometric_shapes', 'hyperbaton', 'logical_deduction_five_objects', 
                      'logical_deduction_seven_objects', 'logical_deduction_three_objects', 'movie_recommendation', 
                      'multistep_arithmetic_two', 'navigate', 'object_counting', 'penguins_in_a_table', 'reasoning_about_colored_objects', 
                      'ruin_names', 'salient_translation_error_detection', 'snarks', 'sports_understanding', 'temporal_sequences', 
                      'tracking_shuffled_objects_five_objects', 'tracking_shuffled_objects_seven_objects', 
                      'tracking_shuffled_objects_three_objects', 'web_of_lies', 'word_sorting']

        for task in self.tasks:
            data = load_dataset("lukaemon/bbh", task)["test"]
            for d in data:
                d["task"] = task
                self.data.append(d)


class DROP(Dataset):
    """
    DROP is a dataset class for the DROP dataset. 
    This dataset is loaded from huggingface datasets: drop (validation set).

    Reference:
    https://huggingface.co/datasets/drop
    DROP: A Reading Comprehension Benchmark Requiring Discrete Reasoning Over Paragraphs (https://arxiv.org/abs/1903.00161)

    Example data format:
    {
        'section_id': 'nfl_1184', 
        'query_id': 'f37e81fa-ef7b-4583-b671-762fc433faa9', 
        'passage': " Hoping to rebound from their loss to the Patriots, the Raiders stayed at home for a Week 16 duel with the Houston Texans.  Oakland would get the early lead in the first quarter as quarterback JaMarcus Russell completed a 20-yard touchdown pass to rookie wide receiver Chaz Schilens.  The Texans would respond with fullback Vonta Leach getting a 1-yard touchdown run, yet the Raiders would answer with kicker Sebastian Janikowski getting a 33-yard and a 30-yard field goal.  Houston would tie the game in the second quarter with kicker Kris Brown getting a 53-yard and a 24-yard field goal. Oakland would take the lead in the third quarter with wide receiver Johnnie Lee Higgins catching a 29-yard touchdown pass from Russell, followed up by an 80-yard punt return for a touchdown.  The Texans tried to rally in the fourth quarter as Brown nailed a 40-yard field goal, yet the Raiders' defense would shut down any possible attempt.", 
        'question': 'Who scored the first touchdown of the game?', 
        'answers_spans': {'spans': ['Chaz Schilens', 'JaMarcus Russell'], 'types': ['span', 'span']}
    }
    """
    def __init__(self):
        data = load_dataset("drop")["validation"]
        self.data = []

        for d in data:
            self.data.append(d)

class ARC(Dataset):
    """
    ARC is a dataset class for the AI2 Reasoning Challenge dataset. 
    This dataset is loaded from huggingface datasets: arc (test set).

    Reference:
    https://huggingface.co/datasets/ai2_arc
    AI2 Reasoning Challenge (ARC) (https://arxiv.org/abs/1803.05457)

    Example data format:
    {
        'id': 'Mercury_7175875', 
        'question': 'An astronomer observes that a planet rotates faster after a meteorite impact. Which is the most likely effect of this increase in rotation?', 
        'choices': {'text': ['Planetary density will decrease.', 'Planetary years will become longer.', 'Planetary days will become shorter.', 'Planetary gravity will become stronger.'], 'label': ['A', 'B', 'C', 'D']}, 
        'answerKey': 'C'
    }
    """
    def __init__(self, name):
        data = load_dataset("ai2_arc", name)["test"]
        self.data = []

        for d in data:
            self.data.append(d)

class VQAv2(Dataset):
    """
    VQAv2 is a dataset class for the Visual Question Answering v2 dataset. 
    This dataset is loaded from huggingface datasets: vqa_v2 (validation set).

    Reference:
    https://huggingface.co/datasets/HuggingFaceM4/VQAv2
    Making the V in VQA Matter: Elevating the Role of Image Understanding in Visual Question Answering (https://arxiv.org/abs/1612.00837)

    Example data format:
    {
        'images': [<PIL.JpegImagePlugin.JpegImageFile image mode=RGB size=640x512>],
        'image_paths': ['/Users/username/.cache/huggingface/datasets/downloads/extracted/adf50b3f63fdc93e9e2d368b11520a5796a4e904a80852e73a19a7112fca3592/val2014/COCO_val2014_000000262148.jpg'],
        'answers': [{'answer': 'down', 'answer_confidence': 'yes', 'answer_id': 1},
        {'answer': 'down', 'answer_confidence': 'yes', 'answer_id': 2},
        {'answer': 'at table', 'answer_confidence': 'yes', 'answer_id': 3},
        {'answer': 'skateboard', 'answer_confidence': 'yes', 'answer_id': 4},
        {'answer': 'down', 'answer_confidence': 'yes', 'answer_id': 5},
        {'answer': 'table', 'answer_confidence': 'yes', 'answer_id': 6},
        {'answer': 'down', 'answer_confidence': 'yes', 'answer_id': 7},
        {'answer': 'down', 'answer_confidence': 'yes', 'answer_id': 8},
        {'answer': 'down', 'answer_confidence': 'yes', 'answer_id': 9},
        {'answer': 'down', 'answer_confidence': 'yes', 'answer_id': 10}],
        'question': 'Where is he looking?'
    }
    """
    def __init__(self):
        from datasets import Image
        data = load_dataset("HuggingFaceM4/VQAv2", split="validation").cast_column("image", Image(decode=False))
        self.data = data
    
    def __getitem__(self, idx):
        assert len(self.data) > 0, "Empty dataset. Please load data first."
        return {"images": [PILImage.open(self.data[idx]['image']['path'])],
                "image_paths": [self.data[idx]['image']['path']],
                "answers": self.data[idx]['answers'],
                "question": self.data[idx]['question'],}

class NoCaps(Dataset):
    """
    NoCaps is a dataset class for the NoCaps dataset. 
    This dataset is loaded from huggingface datasets: nocaps (validation set).

    Reference:
    https://huggingface.co/datasets/HuggingFaceM4/NoCaps
    nocaps: novel object captioning at scale (https://arxiv.org/abs/1812.08658)

    Example data format:
    {
        'images': [<PIL.JpegImagePlugin.JpegImageFile image mode=L size=732x1024>],
        'image_paths': ['https://s3.amazonaws.com/nocaps/val/0013ea2087020901.jpg'],
        'answers': ['A baby is standing in front of a house.',
        'A little girl in a white jacket and sandals.',
        'A young child stands in front of a house.',
        'A child is wearing a white shirt and standing on a side walk. ',
        'A little boy is standing in his diaper with a white shirt on.',
        'A child wearing a diaper and shoes stands on the sidewalk.',
        'A child is wearing a light-colored shirt during the daytime.',
        'A little kid standing on the pavement in a shirt. ',
        'Black and white photo of a little girl smiling.',
        'a cute baby is standing alone with white shirt']
    }
    """
    def __init__(self):
        from datasets import Image
        data = load_dataset("HuggingFaceM4/NoCaps", split="validation").cast_column("image", Image(decode=False))
        data1 = load_dataset("HuggingFaceM4/NoCaps", split="validation")
        self.data = data
        self.data1 = data1
    
    def __getitem__(self, idx):
        assert len(self.data) > 0, "Empty dataset. Please load data first."
        return {"images": [self.data1[idx]['image']],
                "image_paths": [self.data[idx]['image']['path']],
                "answers": self.data[idx]['annotations_captions']}

class MathVista(Dataset):
    """
    MathVista is a dataset class for the MathVista dataset. 
    This dataset is loaded from huggingface datasets: math_vista (testmini set).

    Reference:
    https://huggingface.co/datasets/AI4Math/MathVista
    MathVista: Evaluating Mathematical Reasoning of Foundation Models in Visual Contexts (https://arxiv.org/abs/2310.02255)

    Example data format:
    {
        'images': [<PIL.PngImagePlugin.PngImageFile image mode=RGBA size=1514x720>],
        'image_paths': ['/Users/username/promptbench/promptbench/data/math_vista/testmini/0_decoded_image.png'],
        'answer': '1.2',
        'question': "When a spring does work on an object, we cannot find the work by simply multiplying the spring force by the object's displacement. The reason is that there is no one value for the force-it changes. However, we can split the displacement up into an infinite number of tiny parts and then approximate the force in each as being constant. Integration sums the work done in all those parts. Here we use the generic result of the integration.\r\n\r\nIn Figure, a cumin canister of mass $m=0.40 \\mathrm{~kg}$ slides across a horizontal frictionless counter with speed $v=0.50 \\mathrm{~m} / \\mathrm{s}$. It then runs into and compresses a spring of spring constant $k=750 \\mathrm{~N} / \\mathrm{m}$. When the canister is momentarily stopped by the spring, by what distance $d$ is the spring compressed?\nANSWER TYPE: float"
    }
        
    """
    def __init__(self):
        data = load_dataset("AI4Math/MathVista", split="testmini")
        self.data = data
        self.save_images_to_local(dataset='math_vista', split='testmini', key_list=['decoded_image'])

    def __getitem__(self, idx):
        assert len(self.data) > 0, "Empty dataset. Please load data first."
        return {"images": [self.data[idx]['decoded_image']],
                "image_paths": self.local_path_dict[str(idx)],
                "answer": self.data[idx]['answer'],
                "question":  self.data[idx]['question'] + "\nANSWER TYPE: " + self.data[idx]['answer_type'],}

class AI2D(Dataset):
    """
    AI2D is a dataset class for the AI2D dataset. 
    This dataset is loaded from huggingface datasets: ai2d (test set).

    Reference:
    https://huggingface.co/datasets/lmms-lab/ai2d
    A Diagram Is Worth A Dozen Images (https://arxiv.org/abs/1603.07396)

    Example data format:
    {
        'images': [<PIL.PngImagePlugin.PngImageFile image mode=RGB size=600x449>],
        'image_paths': ['/Users/username/promptbench/promptbench/data/ai2d/test/0_image.png'],
        'question': 'which of these define dairy item\n0: c\n1: D\n2: b\n3: a',
        'answer': '1'
    }
    """
    def __init__(self):
        data = load_dataset("lmms-lab/ai2d", split="test")
        self.data = data
        self.save_images_to_local(dataset='ai2d', split='test', key_list=['image'])

    def __getitem__(self, idx):
        assert len(self.data) > 0, "Empty dataset. Please load data first."
        choices_dict = dict(enumerate(self.data[idx]['options']))
        choices = ''
        for k, v in choices_dict.items():
            choices += f"\n{k}: {v}"
        return({"images": [self.data[idx]['image']],
                "image_paths": self.local_path_dict[str(idx)],
                "question": self.data[idx]['question'] + choices,
                "answer": self.data[idx]['answer']})

class ChartQA(Dataset):
    """
    ChartQA is a dataset class for the ChartQA dataset.
    This dataset is loaded from huggingface datasets: chart_qa (test set).

    Reference:
    https://huggingface.co/datasets/lmms-lab/ChartQA
    ChartQA: A Benchmark for Question Answering about Charts with Visual and Logical Reasoning (https://arxiv.org/abs/2203.10244)

    Example data format:
    {
        'images': [<PIL.PngImagePlugin.PngImageFile image mode=RGBA size=850x600>],
        'image_paths': ['/Users/username/promptbench/promptbench/data/chart_qa/test/0_image.png'],
        'answer': '14',
        'question': 'How many food item is shown in the bar graph?'
    }
    """
    def __init__(self):
        data = load_dataset("lmms-lab/ChartQA", split="test")
        self.data = data
        self.save_images_to_local(dataset='chart_qa', split='test', key_list=['image'])
    
    def __getitem__(self, idx):
        assert len(self.data) > 0, "Empty dataset. Please load data first."
        return {"images": [self.data[idx]['image']],
                "image_paths": self.local_path_dict[str(idx)],
                "answer": self.data[idx]['answer'],
                "question":  self.data[idx]['question'],}

class ScienceQA(Dataset):
    """
    ScienceQA is a dataset class for the ScienceQA dataset.
    This dataset is loaded from huggingface datasets: science_qa (validation set).

    Reference:
    https://huggingface.co/datasets/derek-thomas/ScienceQA
    Learn to Explain: Multimodal Reasoning via Thought Chains for Science Question Answering (https://arxiv.org/abs/2209.09513)

    Example data format:
    {
        'images': [<PIL.PngImagePlugin.PngImageFile image mode=RGB size=202x202>],
        'image_paths': ['/Users/username/promptbench/promptbench/data/science_qa/validation/0_image.png'],
        'answer': 1,
        'question': "Which animal's mouth is also adapted for bottom feeding?\n0: discus\n1: armored catfish"
    }
    """
    def __init__(self):
        data = load_dataset("derek-thomas/ScienceQA", split="validation")
        data_with_images = []
        for d in data:
            if d['image'] is not None:
                data_with_images.append(d)
        self.data = data_with_images
        self.save_images_to_local(dataset='science_qa', split='validation', key_list=['image'])

    def __getitem__(self, idx):
        assert len(self.data) > 0, "Empty dataset. Please load data first."
        choices_dict = dict(enumerate(self.data[idx]['choices']))
        choices = ''
        for k, v in choices_dict.items():
            choices += f"\n{k}: {v}"
        return {"images": [self.data[idx]['image']],
                "image_paths": self.local_path_dict[str(idx)],
                "answer": self.data[idx]['answer'],
                "question":  self.data[idx]['question'] + choices,}

class MMMU(Dataset):
    """
    MMMU is a dataset class for the MMMU dataset. 
    This dataset is loaded from huggingface datasets: mmlu (validation set).

    Reference:
    https://huggingface.co/datasets/lmms-lab/MMMU
    MMMU: A Massive Multi-discipline Multimodal Understanding and Reasoning Benchmark for Expert AGI (https://arxiv.org/abs/2311.16502)

    {
        'images': [<PIL.PngImagePlugin.PngImageFile image mode=RGBA size=733x237>],
        'image_paths': ['/Users/username/promptbench/promptbench/data/mmmu/validation/0_image_1.png'],
        'answer': 'B',
        'question': '<image 1> Baxter Company has a relevant range of production between 15,000 and 30,000 units. The following cost data represents average variable costs per unit for 25,000 units of production. If 30,000 units are produced, what are the per unit manufacturing overhead costs incurred?\nA: $6\nB: $7\nC: $8\nD: $9'
    }
    """
    def __init__(self):
        data = load_dataset("lmms-lab/MMMU", split="validation")
        self.data = data
        self.save_images_to_local(dataset='mmmu', split='validation', key_list=['image_1', 'image_2', 'image_3', 'image_4', 'image_5', 'image_6', 'image_7'])

    def __getitem__(self, idx):
        assert len(self.data) > 0, "Empty dataset. Please load data first."
        
        d = self.data[idx]
        choices_dict = dict(enumerate(eval(d['options'])))
        choices = ''
        for k, v in choices_dict.items():
            choices += f"\n{chr(ord('A') + int(k))}: {v}"
        question = d['question'] + choices

        images = []
        for i in range(1, 7):
            if f'image {i}' in question:
                if d[f'image_{i}'].mode == 'P':
                    d[f'image_{i}'] = d[f'image_{i}'].convert('RGBA')
                images.append(d[f'image_{i}'])
        
        return {"images": images,
                "image_paths": self.local_path_dict[str(idx)],
                "answer": d['answer'],
                "question": question,}



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from abc import ABC
import torch

class LMMBaseModel(ABC):
    """
    Abstract base class for language model interfaces.

    This class provides a common interface for various language models and includes methods for prediction.

    Parameters:
    -----------
    model : str
        The name of the language model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').

    Methods:
    --------
    predict(input_text, **kwargs)
        Generates a prediction based on the input text.
    __call__(input_text, **kwargs)
        Shortcut for predict method.
    """
    def __init__(self, model_name, max_new_tokens, temperature, device='auto'):
        self.model_name = model_name
        self.max_new_tokens = max_new_tokens
        self.temperature = temperature
        self.device = device

    def predict(self, input_text, **kwargs):
        if self.device == 'auto':
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            device = self.device
        input_ids = self.tokenizer(input_text, return_tensors="pt").input_ids.to(device)

        outputs = self.model.generate(input_ids, 
                                     max_new_tokens=self.max_new_tokens, 
                                     temperature=self.temperature,
                                     do_sample=True,
                                     **kwargs)
        
        out = self.tokenizer.decode(outputs[0])
        return out

    def __call__(self, input_text, **kwargs):
        return self.predict(input_text, **kwargs)


class BaichuanModel(LMMBaseModel):
    """
    Language model class for the Baichuan model.

    Inherits from LMMBaseModel and sets up the Baichuan language model for use.

    Parameters:
    -----------
    model : str
        The name of the Baichuan model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float, optional
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').

    Methods:
    --------
    predict(input_text, **kwargs)
        Generates a prediction based on the input text.
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(BaichuanModel, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import AutoTokenizer, AutoModelForCausalLM
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device, use_fast=False, trust_remote_code=True)
        self.model = AutoModelForCausalLM.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device, trust_remote_code=True)


class YiModel(LMMBaseModel):
    """
    Language model class for the Yi model.

    Inherits from LMMBaseModel and sets up the Yi language model for use.

    Parameters:
    -----------
    model : str
        The name of the Yi model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(YiModel, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import AutoTokenizer, AutoModelForCausalLM
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)
        self.model = AutoModelForCausalLM.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)


class MixtralModel(LMMBaseModel):
    """
    Language model class for the Mixtral model.

    Inherits from LMMBaseModel and sets up the Mixtral language model for use.

    Parameters:
    -----------
    model : str
        The name of the Mixtral model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(MixtralModel, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import AutoTokenizer, AutoModelForCausalLM

        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)
        self.model = AutoModelForCausalLM.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)


class MistralModel(LMMBaseModel):
    """
    Language model class for the Mistral model.

    Inherits from LMMBaseModel and sets up the Mistral language model for use.

    Parameters:
    -----------
    model : str
        The name of the Mistral model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(MistralModel, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import AutoTokenizer, AutoModelForCausalLM
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)
        self.model = AutoModelForCausalLM.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)


class PhiModel(LMMBaseModel):
    """
    Language model class for the Phi model.

    Inherits from LMMBaseModel and sets up the Phi language model for use.

    Parameters:
    -----------
    model : str
        The name of the Phi model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(PhiModel, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import AutoTokenizer, AutoModelForCausalLM
        model = "microsoft/phi-1_5" if model_name == "phi-1.5" else "microsoft/phi-2"
        
        self.tokenizer = AutoTokenizer.from_pretrained(model, trust_remote_code=True, torch_dtype=dtype, device_map=device)
        self.model = AutoModelForCausalLM.from_pretrained(model, trust_remote_code=True, torch_dtype=dtype, device_map=device)

    
    def predict(self, input_text, **kwargs):
        if self.device == 'auto':
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            device = self.device
        input_ids = self.tokenizer(input_text, return_tensors="pt").input_ids.to(device)

        outputs = self.model.generate(input_ids, 
                                     max_new_tokens=self.max_new_tokens, 
                                     temperature=self.temperature,
                                     **kwargs)
        
        out = self.tokenizer.decode(outputs[0])
        return out[len(input_text):]

class T5Model(LMMBaseModel):
    """
    Language model class for the T5 model.

    Inherits from LMMBaseModel and sets up the T5 language model for use.

    Parameters:
    -----------
    model : str
        The name of the T5 model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(T5Model, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import T5Tokenizer, T5ForConditionalGeneration
        
        self.tokenizer = T5Tokenizer.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)
        self.model = T5ForConditionalGeneration.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)


class UL2Model(LMMBaseModel):
    """
    Language model class for the UL2 model.

    Inherits from LMMBaseModel and sets up the UL2 language model for use.

    Parameters:
    -----------
    model : str
        The name of the UL2 model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(UL2Model, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import AutoTokenizer, T5ForConditionalGeneration

        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)
        self.model = T5ForConditionalGeneration.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)


class LlamaModel(LMMBaseModel):
    """
    Language model class for the Llama model.

    Inherits from LMMBaseModel and sets up the Llama language model for use.

    Parameters:
    -----------
    model : str
        The name of the Llama model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    system_prompt : str
        The system prompt to be used (default is None).
    model_dir : str
        The directory containing the model files (default is None). If not provided, it will be downloaded from the HuggingFace model hub.
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype, system_prompt, model_dir):
        super(LlamaModel, self).__init__(model_name, max_new_tokens, temperature, device)
        if system_prompt is None:
            self.system_prompt = "You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe.  Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. Please ensure that your responses are socially unbiased and positive in nature.\nIf a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct. If you don't know the answer to a question, please don't share false information."
        else:
            self.system_prompt = system_prompt
        from transformers import AutoTokenizer, AutoModelForCausalLM
        
        if model_dir is None:
            parts = model_name.split('-')
            number = parts[1]
            is_chat = 'chat' in parts

            model_dir = f"meta-llama/Llama-2-{number}"
            if is_chat:
                model_dir += "-chat"
            model_dir += "-hf"
            
        self.tokenizer = AutoTokenizer.from_pretrained(model_dir, device_map=device, torch_dtype=dtype)
        self.model = AutoModelForCausalLM.from_pretrained(model_dir, device_map=device, torch_dtype=dtype)

    def predict(self, input_text, **kwargs):
        if self.device == 'auto':
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            device = self.device

        input_text = f"<s>[INST] <<SYS>>{self.system_prompt}<</SYS>>\n{input_text}[/INST]"
        input_ids = self.tokenizer(input_text, return_tensors="pt").input_ids.to(device)
        
        outputs = self.model.generate(input_ids, 
                                     max_new_tokens=self.max_new_tokens, 
                                     temperature=self.temperature,
                                     **kwargs)
        
        out = self.tokenizer.decode(outputs[0], 
                                    skip_special_tokens=True, 
                                    clean_up_tokenization_spaces=False)
        
        return out[len(input_text):]


class VicunaModel(LMMBaseModel):
    """
    Language model class for the Vicuna model.

    Inherits from LMMBaseModel and sets up the Vicuna language model for use.

    Parameters:
    -----------
    model : str
        The name of the Vicuna model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float, optional
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    model_dir : str, optional
        The directory containing the model files (default is None).
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype, model_dir):
        super(VicunaModel, self).__init__(model_name, max_new_tokens, temperature, device)

        from transformers import AutoModelForCausalLM, AutoTokenizer

        self.tokenizer = AutoTokenizer.from_pretrained(model_dir, device_map=device, torch_dtype=dtype, use_fast=False)
        self.model = AutoModelForCausalLM.from_pretrained(model_dir, device_map=device, torch_dtype=dtype)

    def predict(self, input_text, **kwargs):
        if self.device == 'auto':
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            device = self.device

        input_ids = self.tokenizer(input_text, return_tensors="pt").input_ids.to(device)
        outputs = self.model.generate(input_ids, 
                                     max_new_tokens=self.max_new_tokens,
                                     temperature=self.temperature,
                                     **kwargs)
        
        out = self.tokenizer.decode(outputs[0])
        
        return out[len(input_text):]


class OpenAIModel(LMMBaseModel):
    """
    Language model class for interfacing with OpenAI's GPT models.

    Inherits from LMMBaseModel and sets up a model interface for OpenAI GPT models.

    Parameters:
    -----------
    model : str
        The name of the OpenAI model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    system_prompt : str
        The system prompt to be used (default is None).
    openai_key : str
        The OpenAI API key (default is None).

    Methods:
    --------
    predict(input_text)
        Predicts the output based on the given input text using the OpenAI model.
    """
    def __init__(self, model_name, max_new_tokens, temperature, system_prompt, openai_key):
        super(OpenAIModel, self).__init__(model_name, max_new_tokens, temperature)
        self.openai_key = openai_key
        self.system_prompt = system_prompt

    def predict(self, input_text, **kwargs):
        
        from openai import OpenAI
        client = OpenAI(api_key=self.openai_key)

        if self.system_prompt is None:
            system_messages = {'role': "system", 'content': "You are a helpful assistant."}
        else:
            system_messages = {'role': "system", 'content': self.system_prompt}
        
        if isinstance(input_text, list):
            messages = input_text
        elif isinstance(input_text, dict):
            messages = [input_text]
        else:
            messages = [{"role": "user", "content": input_text}]
        
        messages.insert(0, system_messages)
    
        # extra parameterss
        n = kwargs['n'] if 'n' in kwargs else 1
        temperature = kwargs['temperature'] if 'temperature' in kwargs else self.temperature
        max_new_tokens = kwargs['max_new_tokens'] if 'max_new_tokens' in kwargs else self.max_new_tokens
        
        response = client.chat.completions.create(
            model=self.model_name,
            messages=messages,
            temperature=temperature,
            max_tokens=max_new_tokens,
            n=n,
        )
        
        if n > 1:
            result = [choice.message.content for choice in response.choices]
        else:
            result = response.choices[0].message.content
            
        return result
                

class PaLMModel(LMMBaseModel):
    """
    Language model class for interfacing with PaLM models.

    Inherits from LMMBaseModel and sets up a model interface for PaLM models.

    Parameters:
    -----------
    model : str
        The name of the PaLM model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float, optional
        The temperature for text generation (default is 0).
    api_key : str, optional
        The PaLM API key (default is None).
    """
    def __init__(self, model, max_new_tokens, temperature=0, api_key=None):
        super(PaLMModel, self).__init__(model, max_new_tokens, temperature)
        self.api_key = api_key
    
    def predict(self, input_text, **kwargs):
        import google.generativeai as palm 
        
        palm.configure(api_key=self.api_key)
        models = [m for m in palm.list_models() if 'generateText' in m.supported_generation_methods]
        model = models[0].name
        
        n = kwargs['n'] if 'n' in kwargs else 1
        temperature = kwargs['temperature'] if 'temperature' in kwargs else self.temperature
        max_new_tokens = kwargs['max_new_tokens'] if 'max_new_tokens' in kwargs else self.max_new_tokens
        
        completion = palm.generate_text(
            model=model,
            prompt=input_text,
            temperature=temperature,
            candidate_count = n,
            max_output_tokens=max_new_tokens,
        )
        
        if n > 1:
            result = [cand.output for cand in completion.candidates]
        else:
            result = completion.result
        
        return result
        
class GeminiModel(LMMBaseModel):
    """
    Language model class for interfacing with Google's Gemini models.

    Inherits from LMMBaseModel and sets up a model interface for Gemini models.

    Parameters:
    -----------
    model : str
        The name of the PaLM model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float, optional
        The temperature for text generation (default is 0).
    gemini_key : str, optional
        The Gemini API key (default is None).
    """
    def __init__(self, model, max_new_tokens, temperature=0, gemini_key=None):
        super(GeminiModel, self).__init__(model, max_new_tokens, temperature)
        self.gemini_key = gemini_key
    
    def predict(self, input_text, **kwargs):
        import google.generativeai as genai 
        
        genai.configure(api_key=self.gemini_key)

        # Set up the model
        generation_config = {
        "temperature": self.temperature,
        "top_p": 1,
        "top_k": 1,
        "max_output_tokens": self.max_new_tokens,
        }

        safety_settings = [
        {
            "category": "HARM_CATEGORY_HARASSMENT",
            "threshold": "BLOCK_MEDIUM_AND_ABOVE"
        },
        {
            "category": "HARM_CATEGORY_HATE_SPEECH",
            "threshold": "BLOCK_MEDIUM_AND_ABOVE"
        },
        {
            "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
            "threshold": "BLOCK_MEDIUM_AND_ABOVE"
        },
        {
            "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
            "threshold": "BLOCK_MEDIUM_AND_ABOVE"
        }
        ]

        model = genai.GenerativeModel(model_name="gemini-pro",
                                    generation_config=generation_config,
                                    safety_settings=safety_settings)

        response = model.generate_content(input_text).text

        return response


class VLMBaseModel(ABC):
    """
    Abstract base class for vision language model interfaces.

    This class provides a common interface for various vision language models and includes methods for prediction.

    Parameters:
    -----------
    model : str
        The name of the vision language model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').

    Methods:
    --------
    predict(input_images, input_text, **kwargs)
        Generates a prediction based on the input images and text.
    __call__(input_image, input_text, **kwargs)
        Shortcut for predict method.
    """
    def __init__(self, model_name, max_new_tokens, temperature, device='auto'):
        self.model_name = model_name
        self.max_new_tokens = max_new_tokens
        self.temperature = temperature
        self.device = device
        self.placeholder = ""

    def predict(self, input_images, input_text, **kwargs):
        if self.device == 'auto':
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            device = self.device
                    
        for i in range(len(input_images)):
            input_text = self.placeholder + input_text
        
        if self.enable_multiple_images:
            input_ids = self.processor(text=input_text, images=input_images, return_tensors="pt").to(device)
        else:
            input_ids = self.processor(text=input_text, images=input_images[0], return_tensors="pt").to(device)

        outputs = self.model.generate(**input_ids, 
                                     max_new_tokens=self.max_new_tokens, 
                                     temperature=self.temperature,
                                     do_sample=True,
                                     **kwargs)
        
        out = self.processor.batch_decode(outputs, skip_special_tokens=True)[0]
        return out

    def __call__(self, input_images, input_text, **kwargs):
        return self.predict(input_images, input_text, **kwargs)

class BLIP2Model(VLMBaseModel):
    """
    Vision Language model class for the BLIP2 model.

    Inherits from VLMBaseModel and sets up the BLIP2 vision language model for use.

    Parameters:
    -----------
    model : str
        The name of the BLIP2 model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float, optional
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').

    Parameters of predict method:
    ----------------
    input_images: list of PIL.Image
        The input images.
    input_text: str
        The input text.
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(BLIP2Model, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import Blip2Processor, Blip2ForConditionalGeneration
        self.processor = Blip2Processor.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device, use_fast=False)
        self.model = Blip2ForConditionalGeneration.from_pretrained(self.model_name, torch_dtype=dtype, device_map=device)
        self.enable_multiple_images = False

class LLaVAModel(VLMBaseModel):
    """
    Vision Language model class for the LLaVA model.

    Inherits from VLMBaseModel and sets up the LLaVA vision language model for use.

    Parameters:
    -----------
    model : str
        The name of the LLaVA model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    
    Parameters of predict method:
    ----------------
    input_image: list of PIL.Image
        The input images.
    input_text: str
        The input text. Using <image> as the placeholder for the image.
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(LLaVAModel, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import AutoProcessor, LlavaForConditionalGeneration
        self.processor = AutoProcessor.from_pretrained(self.model_name, device_map=device, trust_remote_code=True)
        self.model = LlavaForConditionalGeneration.from_pretrained(self.model_name, device_map=device)
        self.enable_multiple_images = True
        self.placeholder = "<image>"  # a specialized placeholder of LLaVA model

class GeminiVisionModel(VLMBaseModel):
    """
    Vision Language model class for interfacing with Google's Gemini models.

    Inherits from VLMBaseModel and sets up a model interface for Gemini models.

    Parameters:
    -----------
    model : str
        The name of the PaLM model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float, optional
        The temperature for text generation (default is 0).
    gemini_key : str, optional
        The Gemini API key (default is None).
    
    Parameters of predict method:
    ----------------
    input_image: list of PIL.Image
        The input images.
    input_text: str
        The input text.
    """
    def __init__(self, model, max_new_tokens, temperature, gemini_key=None):
        super(GeminiVisionModel, self).__init__(model, max_new_tokens, temperature)
        self.gemini_key = gemini_key
        self.enable_multiple_images = True
    
    def predict(self, input_images, input_text, **kwargs):
        import google.generativeai as genai 
        
        genai.configure(api_key=self.gemini_key)

        # Set up the model
        generation_config = {
            "temperature": self.temperature,
            "top_p": 1,
            "top_k": 1,
            "max_output_tokens": self.max_new_tokens,
        }

        safety_settings = [
            {
                "category": "HARM_CATEGORY_HARASSMENT",
                "threshold": "BLOCK_MEDIUM_AND_ABOVE"
            },
            {
                "category": "HARM_CATEGORY_HATE_SPEECH",
                "threshold": "BLOCK_MEDIUM_AND_ABOVE"
            },
            {
                "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
                "threshold": "BLOCK_MEDIUM_AND_ABOVE"
            },
            {
                "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
                "threshold": "BLOCK_MEDIUM_AND_ABOVE"
            }
        ]

        model = genai.GenerativeModel(model_name="gemini-pro-vision",
                                      generation_config=generation_config,
                                      safety_settings=safety_settings)

        response = model.generate_content(input_images + [input_text])

        try:
            return response.text
        except:
            print('Error when generating the response using Gemini model')
            return ""

class OpenAIVisionModel(VLMBaseModel):
    """
    Vision Language model class for interfacing with OpenAI's GPT models.

    Inherits from VLMBaseModel and sets up a model interface for OpenAI GPT models.

    Parameters:
    -----------
    model : str
        The name of the OpenAI model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    system_prompt : str
        The system prompt to be used (default is None).
    openai_key : str
        The OpenAI API key (default is None).

    Parameters of predict method:
    ----------------
    input_image: list of str
        The url / local path of the input images.
    input_text: str
        The input text.
    """
    def __init__(self, model_name, max_new_tokens, temperature, system_prompt, openai_key):
        super(OpenAIVisionModel, self).__init__(model_name, max_new_tokens, temperature)
        self.openai_key = openai_key
        self.system_prompt = system_prompt
        self.enable_multiple_images = True

    def predict(self, input_images, input_text, **kwargs):

        if self.system_prompt is None:
            system_messages = {'role': "system", 'content': "You are a helpful assistant."}
        else:
            system_messages = {'role': "system", 'content': self.system_prompt}
        # extra parameterss
        n = kwargs['n'] if 'n' in kwargs else 1
        temperature = kwargs['temperature'] if 'temperature' in kwargs else self.temperature
        max_new_tokens = kwargs['max_new_tokens'] if 'max_new_tokens' in kwargs else self.max_new_tokens
        
        # for input image with url
        if "http://" in input_images[0] or "https://" in input_images[0]:
        
            from openai import OpenAI
            client = OpenAI(api_key=self.openai_key)

            messages = [{"role": "user", 
                         "content": [
                             {"type": "text", "text": input_text},
                         ]}]
            messages.insert(0, system_messages)
            for input_image in input_images:
                messages[1]['content'].append({
                                 "type": "image_url",
                                 "image_url": {"url": input_image},
                             })
                
            response = client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                temperature=temperature,
                max_tokens=max_new_tokens,
                n=n,
            )
            
            if n > 1:
                result = [choice.message.content for choice in response.choices]
            else:
                result = response.choices[0].message.content
                
            return result
        
        # for input image with local path
        else:
            import base64
            import requests

            api_key = self.openai_key

            def encode_image(image_path):
                with open(image_path, "rb") as image_file:
                    return base64.b64encode(image_file.read()).decode('utf-8')

            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {api_key}"
            }

            payload = {
                "model": "gpt-4-vision-preview",
                "messages": [
                    system_messages, 
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "text",
                                "text": input_text
                            },
                        ]
                    }
                ],
                "temperature": temperature,
                "max_tokens": max_new_tokens,
                "n": n,
            }
            base64_images = []
            for input_image in input_images:
                base64_image = encode_image(input_image)  # Getting the base64 string
                base64_images.append(base64_image)
            for base64_img in base64_images:
                payload['messages'][1]['content'].append({
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:image/jpeg;base64, {base64_img}"
                    }
                })

            response = requests.post("https://api.openai.com/v1/chat/completions", headers=headers, json=payload)

            if n > 1:
                result = [choice['message']['content'] for choice in response.json()['choices']]
            else:
                result = response.json()['choices'][0]['message']['content']

            return result
    
class QwenVLModel(VLMBaseModel):
    """
    Vision Language model class for the Qwen model.

    Inherits from VLMBaseModel and sets up the Qwen vision language model for use.

    Parameters:
    -----------
    model : str
        The name of the Qwen model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    system_prompt : str
        The system prompt to be used (default is None).
    api_key : str
        The api key for the Qwen model (default is None).
    
    Parameters of predict method:
    ----------------
    input_image: list of str
        The url / local path of the input images.
        (Add "file://" prefix for local path when using 'qwen-vl-plus' and 'qwen-vl-max')  
    input_text: str
        The input text.
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype, system_prompt, api_key):
        if model_name in ['qwen-vl-plus', 'qwen-vl-max']:
            super(QwenVLModel, self).__init__(model_name, max_new_tokens, temperature)
            assert api_key is not None, f"API key is required for {model_name}"
            self.api_key = api_key
            self.system_prompt = system_prompt
        else:
            super(QwenVLModel, self).__init__(model_name, max_new_tokens, temperature, device)
            from transformers import AutoModelForCausalLM, AutoTokenizer
            self.model = AutoModelForCausalLM.from_pretrained(model_name, torch_dtype=dtype, device_map=device, trust_remote_code=True).eval()
            self.tokenizer = AutoTokenizer.from_pretrained(model_name, torch_dtype=dtype, device_map=device, trust_remote_code=True)
        self.enable_multiple_images = True

    def predict(self, input_images, input_text, **kwargs):

        if self.model_name in ['qwen-vl-plus', 'qwen-vl-max']:
            from http import HTTPStatus
            import dashscope
            dashscope.api_key = self.api_key
            if self.system_prompt is None:
                system_messages = {
                    'role': 'system',
                    'content': [{
                        'text': 'You are a helpful assistant.'
                    }]
                }
            else:
                system_messages = {
                    'role': 'system',
                    'content': [{
                        'text': self.system_prompt
                    }]
                }
            messages = [{
                'role': 'user',
                'content': [{'image': input_image} for input_image in input_images] + [{'text': input_text}]
            }]
            messages.insert(0, system_messages)
            response = dashscope.MultiModalConversation.call(model=self.model_name,
                                                            messages=messages)

            if response.status_code == HTTPStatus.OK:
                return response['output']['choices'][0]['message']['content'][0]['text']
            else:
                print(response.code)  # The error code.
                print(response.message)  # The error message.
                return ""

        else:
            query = self.tokenizer.from_list_format(
                [{'image': input_image} for input_image in input_images] + [{'text': input_text}]
            )
            response, _ = self.model.chat(self.tokenizer, query=query, history=None,
                                          max_new_tokens=self.max_new_tokens, temperature=self.temperature)
            return response

class InternLMVisionModel(VLMBaseModel):
    """
    Vision Language model class for interfacing with InternLM's vision language models.

    Inherits from VLMBaseModel and sets up a model interface for InternLM's vision language models.

    Parameters:
    -----------
    model_name : str
        The name of the InternLM model.
    max_new_tokens : int
        The maximum number of new tokens to be generated.
    temperature : float, optional
        The temperature for text generation (default is 0).
    device: str
        The device to use for inference (default is 'auto').
    dtype: str
        The dtype to use for inference (default is 'auto').
    
    Parameters of predict method:
    ----------------
    input_image: list of str
        The url / local path of the input images.
    input_text: str
        The input text.
    """
    def __init__(self, model_name, max_new_tokens, temperature, device, dtype):
        super(InternLMVisionModel, self).__init__(model_name, max_new_tokens, temperature, device)
        from transformers import AutoModel, AutoTokenizer
        self.model = AutoModel.from_pretrained(model_name, torch_dtype=dtype, device_map=device, trust_remote_code=True).eval()
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, torch_dtype=dtype, device_map=device, trust_remote_code=True)
        self.enable_multiple_images = False
        self.placeholder = "<ImageHere>"  # a specialized placeholder of InternLM model
    
    def predict(self, input_images, input_text, **kwargs):
        input_text = self.placeholder + input_text
        with torch.cuda.amp.autocast():
            response, _ = self.model.chat(self.tokenizer, query=input_text, image=input_images[0], history=[], do_sample=True,
                                          max_new_tokens=self.max_new_tokens, temperature=self.temperature)
        return response


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from .dataset import *

SUPPORTED_DATASETS = [
    "sst2", "cola", "qqp",
    "mnli", "mnli_matched", "mnli_mismatched",
    "qnli", "wnli", "rte", "mrpc",
    "mmlu", "squad_v2", "un_multi", "iwslt2017", "math",
    "bool_logic", "valid_parentheses",
    "gsm8k", "csqa", "bigbench_date", "bigbench_object_tracking",
    "last_letter_concat", "numersense", "qasc",
    "bbh", "drop", "arc-easy", "arc-challenge",
]

SUPPORTED_DATASETS_VLM = [
    "vqav2", "nocaps", "science_qa", 
    "math_vista", "ai2d", "mmmu", "chart_qa"
]

class DatasetLoader:
    
    @staticmethod
    def load_dataset(dataset_name, task=None, supported_languages=None):
        """
        Load and return the specified dataset.

        This function acts as a factory method, returning the appropriate dataset object 
        based on the provided dataset name. 
        'math', 'un_multi' and 'iwslt' require additional arguments to specify the languages used in the dataset.

        Args:
            dataset_name (str): The name of the dataset to load.
            task: str: Additional arguments required by 'math'. 
                        Please visit https://huggingface.co/datasets/math_dataset/ to see the supported tasks for math.
            supported_languages: list: Additional arguments required by 'iwslt'. 
                                Please visit https://huggingface.co/datasets/iwslt2017 to see the supported languages for iwslt.
                                e.g. supported_languages=['de-en', 'ar-en'] for German-English and Arabic-English translation.
        Returns:
            Dataset object corresponding to the given dataset_name.
            The dataset object is an instance of a list, each element is a dictionary. Please refer to each dataset's documentation for details.

        Raises:
            NotImplementedError: If the dataset_name does not correspond to any known dataset.
        """        
        # GLUE datasets
        if dataset_name in ["cola", "sst2", "qqp", "mnli", "mnli_matched", "mnli_mismatched", 
                            "qnli", "wnli", "rte", "mrpc"]:
            return GLUE(dataset_name)
        elif dataset_name == 'mmlu':
            return MMLU()
        elif dataset_name == "squad_v2":
            return SQUAD_V2()
        elif dataset_name == 'un_multi':
            return UnMulti()
        elif dataset_name == 'iwslt2017':
            return IWSLT(supported_languages)
        elif dataset_name in 'math':
            return Math(task)
        elif dataset_name == 'bool_logic':
            return BoolLogic()
        elif dataset_name == 'valid_parentheses':
            return ValidParentheses()
        elif dataset_name == 'gsm8k':
            return GSM8K()
        elif dataset_name == 'csqa':
            return CSQA()
        elif dataset_name == 'qasc':
            return QASC()
        elif 'bigbench' in dataset_name:
            return BigBench(dataset_name)
        elif dataset_name == 'bbh':
            return BBH()
        elif dataset_name == 'drop':
            return DROP()
        elif dataset_name == 'arc-easy':
            return ARC('ARC-Easy')
        elif dataset_name == 'arc-challenge':
            return ARC('ARC-Challenge')
        elif dataset_name == 'vqav2':
            return VQAv2()
        elif dataset_name =='nocaps':
            return NoCaps()
        elif dataset_name =='math_vista':
            return MathVista()
        elif dataset_name == 'ai2d':
            return AI2D()
        elif dataset_name == 'mmmu':
            return MMMU()
        elif dataset_name == 'chart_qa':
            return ChartQA()
        elif dataset_name == 'science_qa':
            return ScienceQA()
        else:
            # If the dataset name doesn't match any known datasets, raise an error
            raise NotImplementedError(f"Dataset '{dataset_name}' is not supported.")




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from .models import *

# A dictionary mapping of model architecture to its supported model names
MODEL_LIST = {
    T5Model: ['google/flan-t5-large'],
    LlamaModel: ['llama2-7b', 'llama2-7b-chat', 'llama2-13b', 'llama2-13b-chat', 'llama2-70b', 'llama2-70b-chat',],
    PhiModel: ['phi-1.5', 'phi-2'],
    PaLMModel: ['palm'],
    OpenAIModel: ['gpt-3.5-turbo', 'gpt-4', 'gpt-4-1106-preview', 'gpt-3.5-turbo-1106', 'gpt-4-0125-preview', 'gpt-3.5-turbo-0125'],
    VicunaModel: ['vicuna-7b', 'vicuna-13b', 'vicuna-13b-v1.3'],
    UL2Model: ['google/flan-ul2'],
    GeminiModel: ['gemini-pro'],
    MistralModel: ['mistralai/Mistral-7B-v0.1', 'mistralai/Mistral-7B-Instruct-v0.1'],
    MixtralModel: ['mistralai/Mixtral-8x7B-v0.1', 'mistralai/Mixtral-8x7B-Instruct-v0.1'],
    YiModel: ['01-ai/Yi-6B', '01-ai/Yi-34B', '01-ai/Yi-6B-Chat', '01-ai/Yi-34B-Chat'],
    BaichuanModel: ['baichuan-inc/Baichuan2-7B-Base', 'baichuan-inc/Baichuan2-13B-Base',
                    'baichuan-inc/Baichuan2-7B-Chat', 'baichuan-inc/Baichuan2-13B-Chat'],
}
MODEL_LIST_VLM = {
    BLIP2Model: ['Salesforce/blip2-opt-2.7b', 'Salesforce/blip2-opt-6.7b',
                 'Salesforce/blip2-flan-t5-xl', 'Salesforce/blip2-flan-t5-xxl'],
    LLaVAModel: ['llava-hf/llava-1.5-7b-hf', 'llava-hf/llava-1.5-13b-hf'],
    GeminiVisionModel: ['gemini-pro-vision'],
    OpenAIVisionModel: ['gpt-4-vision-preview'],
    QwenVLModel: ['Qwen/Qwen-VL', 'Qwen/Qwen-VL-Chat',
                  'qwen-vl-plus', 'qwen-vl-max'],
    InternLMVisionModel: ['internlm/internlm-xcomposer2-vl-7b'],
}

SUPPORTED_MODELS = [model for model_class in MODEL_LIST.keys() for model in MODEL_LIST[model_class]]
SUPPORTED_MODELS_VLM = [model for model_class in MODEL_LIST_VLM.keys() for model in MODEL_LIST_VLM[model_class]]


class LLMModel(object):
    """
    A class providing an interface for various language models.

    This class supports creating and interfacing with different language models, handling prompt engineering, and performing model inference.

    Parameters:
    -----------
    model : str
        The name of the model to be used.
    max_new_tokens : int, optional
        The maximum number of new tokens to be generated (default is 20).
    temperature : float, optional
        The temperature for text generation (default is 0).
    device : str, optional
        The device to be used for inference (default is "cuda").
    dtype : str, optional
        The loaded data type of the language model (default is "auto").
    model_dir : str or None, optional
        The directory containing the model files (default is None).
    system_prompt : str or None, optional
        The system prompt to be used (default is None).
    api_key : str or None, optional
        The API key for API-based models (GPT series and Gemini series), if required (default is None).

    Methods:
    --------
    _create_model(max_new_tokens, temperature, device, dtype, model_dir, system_prompt, api_key)
        Creates and returns the appropriate model instance.
    convert_text_to_prompt(text, role)
        Constructs a prompt based on the text and role.
    concat_prompts(prompt_list)
        Concatenates multiple prompts into a single prompt.
    _gpt_concat_prompts(prompt_list)
        Concatenates prompts for GPT models.
    _other_concat_prompts(prompt_list)
        Concatenates prompts for non-GPT models.
    __call__(input_text, **kwargs)
        Makes a prediction based on the input text using the loaded model.
    """
    
    @staticmethod
    def model_list():
        return SUPPORTED_MODELS

    def __init__(self, model: str, max_new_tokens: int=20, temperature: float=0.0, device: str="cuda", dtype: str="auto", model_dir: str=None, system_prompt: str=None, api_key:str =None):
        self.model_name = model
        self.model = self._create_model(max_new_tokens, temperature, device, dtype, model_dir, system_prompt, api_key)

    def _create_model(self, max_new_tokens, temperature, device, dtype, model_dir, system_prompt, api_key):
        """Creates and returns the appropriate model based on the model name."""

        # Dictionary mapping of model names to their respective classes
        model_mapping = {model: model_class for model_class in MODEL_LIST.keys() for model in MODEL_LIST[model_class]}

        # Get the model class based on the model name and instantiate it
        model_class = model_mapping.get(self.model_name)
        if model_class:
            if model_class == LlamaModel or model_class == VicunaModel:
                return model_class(self.model_name, max_new_tokens, temperature, device, dtype, system_prompt, model_dir)
            elif model_class in [OpenAIModel]:
                return model_class(self.model_name, max_new_tokens, temperature, system_prompt, api_key)
            elif model_class in [PaLMModel, GeminiModel]:
                return model_class(self.model_name, max_new_tokens, temperature, api_key)
            else:
                return model_class(self.model_name, max_new_tokens, temperature, device, dtype)
        else:
            raise ValueError("The model is not supported!")

    
    def convert_text_to_prompt(self, text, role):
        """Constructs multi_turn conversation for complex methods in prompt engineering."""
        if self.model_name == ['gpt4', 'gpt-3.5-turbo-16k', 'gpt-3.5-turbo-0301', 'gpt-3.5-turbo']:
            return {'role': role, 'content': text} 
        else:
            # return str(role) + ': ' + str(text) + '\n'
            return str(text) + '\n'

    def concat_prompts(self, prompt_list):
        """Concatenates the prompts into a single prompt."""
        if self.model_name == ['gpt4', 'gpt-3.5-turbo-16k', 'gpt-3.5-turbo-0301', 'gpt-3.5-turbo']:
            return self._gpt_concat_prompts(prompt_list)
        else:
            return self._other_concat_prompts(prompt_list)
        
    def _gpt_concat_prompts(self, prompt_list):
        """
        Concatenate prompts from various inputs into a single list of dictionaries.

        The function accepts any number of keyword arguments, each of which can be either
        a dictionary or a list of dictionaries. It concatenates all inputs into a single list.

        Returns:
            A list of dictionaries containing all the prompts from the input arguments.
        """
        # Initialize an empty list to hold all dictionaries
        all_prompts = []

        # Iterate over each keyword argument
        for arg in prompt_list:
            # Check if the argument is a dictionary, and if so, add it to the list
            if isinstance(arg, dict):
                all_prompts.append(arg)
            # Check if the argument is a list of dictionaries
            elif isinstance(arg, list) and all(isinstance(item, dict) for item in arg):
                # Extend the list with the dictionaries from the current argument
                all_prompts.extend(arg)
            else:
                raise ValueError("All arguments must be dictionaries or lists of dictionaries.")

        return all_prompts

    def _other_concat_prompts(self, prompt_list):
        """
        Concatenate prompts from various inputs into a single strings.

        The function accepts any number of keyword arguments, each of which must be
        a string. It concatenates all inputs into a single string.

        Returns:
            A string containing all the prompts from the input arguments.
        """
        # Initialize an empty string to hold all prompts
        all_prompts = ""

        # Iterate over each keyword argument
        for arg in prompt_list:
            # Check if the argument is a string, and if so, add it to the list
            if isinstance(arg, str):
                all_prompts = all_prompts + '\n' + arg
            else:
                raise ValueError("All arguments must be strings.")

        return all_prompts
    
    def __call__(self, input_text, **kwargs):
        """Predicts the output based on the given input text using the loaded model."""
        return self.model.predict(input_text, **kwargs)


class VLMModel(object):
    """
    A class providing an interface for various vision language models.

    This class supports creating and interfacing with different vision language models, handling prompt engineering, and performing model inference.

    Parameters:
    -----------
    model : str
        The name of the model to be used.
    max_new_tokens : int, optional
        The maximum number of new tokens to be generated (default is 20).
    temperature : float, optional
        The temperature for text generation (default is 0).
    device : str, optional
        The device to be used for inference (default is "cuda").
    dtype : str, optional
        The loaded data type of the language model (default is "auto").
    model_dir : str or None, optional
        The directory containing the model files (default is None).
    system_prompt : str or None, optional
        The system prompt to be used (default is None).
    api_key : str or None, optional
        The API key for API-based models (GPT series, Gemini series and Qwen series), if required (default is None).

    Methods:
    --------
    _create_model(max_new_tokens, temperature, device, dtype, model_dir, system_prompt, api_key)
        Creates and returns the appropriate model instance.
    convert_text_to_prompt(text, role)
        Constructs a prompt based on the text and role.
    concat_prompts(prompt_list)
        Concatenates multiple prompts into a single prompt.
    _gpt_concat_prompts(prompt_list)
        Concatenates prompts for GPT models.
    _other_concat_prompts(prompt_list)
        Concatenates prompts for non-GPT models.
    __call__(input_images, input_text, **kwargs)
        Makes a prediction based on the input images and input text using the loaded model.
    """
    
    @staticmethod
    def model_list():
        return SUPPORTED_MODELS_VLM

    def __init__(self, model: str, max_new_tokens: int=20, temperature: float=0.0, device: str="cuda", dtype: str="auto", system_prompt: str=None, api_key:str =None):
        self.model_name = model
        self.model = self._create_model(max_new_tokens, temperature, device, dtype, system_prompt, api_key)

    def _create_model(self, max_new_tokens, temperature, device, dtype, system_prompt, api_key):
        """Creates and returns the appropriate model based on the model name."""

        # Dictionary mapping of model names to their respective classes
        model_mapping = {model: model_class for model_class in MODEL_LIST_VLM.keys() for model in MODEL_LIST_VLM[model_class]}

        # Get the model class based on the model name and instantiate it
        model_class = model_mapping.get(self.model_name)
        if model_class:
            if model_class in [OpenAIVisionModel]:
                return model_class(self.model_name, max_new_tokens, temperature, system_prompt, api_key)
            elif model_class in [GeminiVisionModel]:
                return model_class(self.model_name, max_new_tokens, temperature, api_key)
            elif model_class in [QwenVLModel]:
                return model_class(self.model_name, max_new_tokens, temperature, device, dtype, system_prompt, api_key)
            else:
                return model_class(self.model_name, max_new_tokens, temperature, device, dtype)
        else:
            raise ValueError("The model is not supported!")

    def convert_text_to_prompt(self, text, role):
        """Constructs multi_turn conversation for complex methods in prompt engineering."""
        if self.model_name == ['gpt-4-vision-preview']:
            return {'role': role, 'content': text} 
        else:
            # return str(role) + ': ' + str(text) + '\n'
            return str(text) + '\n'

    def concat_prompts(self, prompt_list):
        """Concatenates the prompts into a single prompt."""
        if self.model_name == ['gpt-4-vision-preview']:
            return self._gpt_concat_prompts(prompt_list)
        else:
            return self._other_concat_prompts(prompt_list)
        
    def _gpt_concat_prompts(self, prompt_list):
        """
        Concatenate prompts from various inputs into a single list of dictionaries.

        The function accepts any number of keyword arguments, each of which can be either
        a dictionary or a list of dictionaries. It concatenates all inputs into a single list.

        Returns:
            A list of dictionaries containing all the prompts from the input arguments.
        """
        # Initialize an empty list to hold all dictionaries
        all_prompts = []

        # Iterate over each keyword argument
        for arg in prompt_list:
            # Check if the argument is a dictionary, and if so, add it to the list
            if isinstance(arg, dict):
                all_prompts.append(arg)
            # Check if the argument is a list of dictionaries
            elif isinstance(arg, list) and all(isinstance(item, dict) for item in arg):
                # Extend the list with the dictionaries from the current argument
                all_prompts.extend(arg)
            else:
                raise ValueError("All arguments must be dictionaries or lists of dictionaries.")

        return all_prompts

    def _other_concat_prompts(self, prompt_list):
        """
        Concatenate prompts from various inputs into a single strings.

        The function accepts any number of keyword arguments, each of which must be
        a string. It concatenates all inputs into a single string.

        Returns:
            A string containing all the prompts from the input arguments.
        """
        # Initialize an empty string to hold all prompts
        all_prompts = ""

        # Iterate over each keyword argument
        for arg in prompt_list:
            # Check if the argument is a string, and if so, add it to the list
            if isinstance(arg, str):
                all_prompts = all_prompts + '\n' + arg
            else:
                raise ValueError("All arguments must be strings.")

        return all_prompts
    
    def __call__(self, input_images, input_text, **kwargs):
        """Predicts the output based on the given input text using the loaded model."""
        return self.model.predict(input_images, input_text, **kwargs)